#!/usr/bin/env python3
"""
NCBI CBB REST client (Computational Biology Branch)
/CBBresearch/Lu/Demo/RESTful/tmTool.cgi/Disease/19894120/JSON
"""
### 
import sys,os,re,json,argparse,time,logging
#
from .. import rest
#
API_HOST='www.ncbi.nlm.nih.gov'
API_BASE_PATH='/CBBresearch/Lu/Demo/RESTful/tmTool.cgi'
#
OFMTS={'XML':'application/xml','JSON':'application/json'}
#
HEADERS={'Accept':OFMTS['JSON']}
#
#
##############################################################################
def GetLinks(pmids, bioconcept, base_url, fout):
  '''One TSV row per unique str.  Count multiple occurances.'''
  n_out=0;
  fout.write('pmid\tsourcedb\tbioconcept\ttext\tcount\n')
  for pmid in pmids:
    try:
      rval=rest.GetURL(base_url+'/%s/%d/JSON'%(bioconcept, pmid), headers=HEADERS, parse_json=True)
    except Exception as e:
      logging.error('%s'%(e))
      continue
    if not type(rval) is dict:
      logging.info('%s'%(str(rval)))
      continue

    sourceid = rval['sourceid'] if 'sourceid' in rval else None
    if not sourceid or pmid!=int(sourceid):
      logging.error('(%s;%d) PMID != sourceid (%s)'%(bioconcept, pmid, sourceid))
      continue
    dens = rval['denotations'] if 'denotations' in rval else []
    source = rval['sourcedb'] if 'sourcedb' in rval else ''
    text = rval['text'] if 'text' in rval else ''
    strs={};
    for den in dens:
      obj = den['obj'] if 'obj' in den else None
      span = den['span'] if 'span' in den and 'begin' in den['span'] and 'end' in den['span'] else None
      if obj and obj.startswith(bioconcept):
        if not text.strip() or not span or span['begin']>=len(text) or  span['end']>len(text):
          logging.error('(%s;%d) span [%s] out of range for text: "%s"'%(bioconcept, pmid, str(span), text))
          continue
        s=text[(span['begin']):span['end']]
        if s in strs:
          if obj in strs[s]:
            strs[s][obj]+=1
          else:
            strs[s][obj]=1
        else:
          strs[s]={obj:1}
    for s in sorted(strs.keys()):
      for obj in sorted(strs[s].keys()):
        fout.write('%d\t"%s"\t"%s"\t"%s"\t%d\n'%(pmid, source, obj, s, strs[s][obj]))
        n_out+=1
    logging.debug((json.dumps(rval, indent=2, sort_keys=False)))
  logging.info('(%s) PMIDs in: %d ; records out: %d'%(bioconcept, len(pmids), n_out))

##############################################################################
if __name__=='__main__':
  BIOCONCEPTS=['Disease','Gene','Chemical','Species','Mutation']
  parser = argparse.ArgumentParser(description='NCBI CBB Bioconcepts NER in PubMed docs REST API client', epilog='')
  ops = ['getlinks']
  parser.add_argument("op", choices=ops, help='operation')
  parser.add_argument("--ids", help="query PubMed IDs (comma-separated)")
  parser.add_argument("--i", dest="ifile", help="input query PubMed IDs")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--bioconcept", choices=BIOCONCEPTS, default="Disease", help="")
  parser.add_argument("--api_host", default=API_HOST)
  parser.add_argument("--api_base_path", default=API_BASE_PATH)
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  base_url='https://'+args.api_host+args.api_base_path

  if args.ofile:
    fout=open(args.ofile,"w")
    if not fout: parser.error('cannot open outfile: %s'%args.ofile)
  else:
    fout=sys.stdout

  t0=time.time()

  if args.ifile:
    fin=open(args.ifile)
    if not fin: parser.error('cannot open pmidfile: %s'%args.ifile)
    pmids=[];
    while True:
      line=fin.readline()
      if not line: break
      try:
        pmid=int(line.rstrip())
      except Exception as e:
        logging.error('invalid PMID: "%s"'%(line.rstrip()))
        continue
      if line.rstrip(): pmids.append(pmid)
    logging.info('%s: input queries: %d'%(PROG, len(pmids)))
    fin.close()
  elif args.ids:
    pmids = re.split('[,\s]+', args.ids.strip())

  if args.op == "getlinks":
    GetLinks(ids, args.bioconcept, base_url, fout)

  else:
    parser.error("Invalid operation: %s"%args.op)

  logging.info(('elapsed time: %s'%(time.strftime('%Hh:%Mm:%Ss',time.gmtime(time.time()-t0)))))



#!/usr/bin/env python3
"""
https://igraph.org/python/doc/tutorial/tutorial.html

Note that igraph defines its own IDs.  These are integers, not the same
as imported GraphML "id" values.

"""
#############################################################################
import sys,os,argparse,logging,tqdm
import re,random,tempfile,shutil,json
import numpy as np
import pandas as pd
import igraph

#############################################################################
def Load_GraphML(ifile):
  g = igraph.Graph.Read_GraphML(ifile)
  logging.info(f"nodes: {g.vcount()}; edges: {g.ecount()}")
  return g

#############################################################################
def Save_GraphML(g, fout):
  g.write_graphml(fout) #Works but maybe changes tags?

#############################################################################
def GraphSummary(g):
  #igraph.summary(g,verbosity=0) ## verbosity=1 prints edge list!
  name = g['name'] if 'name' in g.attributes() else None
  logging.info(f"graph name: '{name}'")
  logging.info(f"                 nodes: {g.vcount():3d}")
  logging.info(f"                 edges: {g.ecount():3d}")
  logging.info(f"             connected: {g.is_connected(mode=igraph.WEAK)}")
  logging.info(f"            components: {len(g.components(mode=igraph.WEAK)):3d}")
  logging.info(f"              directed: {g.is_directed()}")
  logging.info(f"DAG (directed-acyclic): {g.is_dag()}")
  logging.info(f"              weighted: {g.is_weighted()}")
  logging.info(f"              diameter: {g.diameter():3d}")
  logging.info(f"                radius: {g.radius():.3f}")
  logging.info(f"             maxdegree: {g.maxdegree():3d}")

#############################################################################
def XOR(a,b): return ((a and not b) or (b and not a))

#############################################################################
def NodeSelect_String(g,selectfield,selectquery,exact,negate):
  vs=[]
  if exact:
    vs = eval(f"g.vs.select({selectfield}_{'ne' if negate else 'eq'} = '{selectquery}')")
  else:
    for v in g.vs:
      if XOR(re.search(selectquery, v[selectfield], re.I), negate):
        vs.append(v)
  return vs

#############################################################################
def ConnectedNodes(g):
  vs=[]
  for v in g.vs:
    #if g.neighbors(v, igraph.ALL):
    if v.degree()>0:
      vs.append(v)
  return vs

#############################################################################
def DisconnectedNodes(g):
  vs=[]
  for v in g.vs:
    #if not g.neighbors(v, igraph.ALL):
    if v.degree()==0:
      vs.append(v)
  return vs

#############################################################################
def RootNodes(g):
  """In a directed graph, which are the root nodes?"""
  if not g.is_directed():
    logging.error("Graph not directed; cannot have root nodes.")
    return []
  elif not g.is_dag():
    logging.error("Graph not directed-acyclic; cannot have proper root nodes.")
    return []
  vs=[];
  for v in g.vs:
    #if not g.neighbors(v, igraph.IN):
    #if g.indegree(v)==0:
    if v.indegree()==0:
      vs.append(v)
  return vs

#############################################################################
def AddChildren(vs, r, depth=1):
  if depth<=0: return
  for c in r.neighbors(igraph.OUT):
    vs.append(c)
    AddChildren(vs, c, depth-1)

#############################################################################
def ChildCount(r, depth=1):
  if depth<=0: return 0
  count=0;
  for c in r.neighbors(igraph.OUT):
    count+=(1+ChildCount(c, depth-1))
  return count

#############################################################################
def AllChildCount(r):
  count=0;
  for c in r.neighbors(igraph.OUT):
    count+=(1+AllChildCount(c))
  return count

#############################################################################
def LeafChildCount(r):
  if len(r.neighbors(igraph.OUT))==0:
    count=1
  else:
    count=0;
    for c in r.neighbors(igraph.OUT):
      count+=LeafChildCount(c)
  return count

#############################################################################
def TopNodes(g, depth=1):
  """Return top node[s] of DAG to given depth from root[s]."""
  vs=[];
  rs = RootNodes(g)
  vs.extend(rs)
  for r in rs:
    Hierarchy(r, depth, 0)
    AddChildren(vs, r, depth)
  logging.info(f"Returned {len(rs)} root node[s] and hierarchy to depth {depth}")
  return vs

#############################################################################
def Hierarchy(n, depth, i):
  if i>depth: return
  indent = "".join([f" {ii}> " for ii in range(i+1)])
  n_child = ChildCount(n, 1)
  n_all = AllChildCount(n)
  n_leaf = LeafChildCount(n)
  logging.info(f"{indent}{n['id']:>12}: {n['name']} (child:{n_child}; all:{n_all}; leaf:{n_leaf})")
  for c in n.neighbors(igraph.OUT):
    Hierarchy(c, depth, i+1)

#############################################################################
def DegreeDistribution(g):
  dd = g.degree_distribution(bin_width=1, mode=igraph.ALL, loops=True)
  logging.info(f"{dd}")

#############################################################################
def ShortestPaths(g, nidA, nidB):
  """Must use GraphBase (not Graph) method to get path data."""
  vs=set();
  vA = g.vs.find(id = nidA)
  vB = g.vs.find(id = nidB)
  paths = g.get_shortest_paths(vA, [vB], weights=None, mode=igraph.ALL, output="vpath")
  for path in paths:
    logging.debug(f"path = {str(path)}")
    n_v=len(path)
    for j in range(n_v):
      vid=path[j]
      v = g.vs[vid]
      vs.add(v)
      dr=None
      if j+1<n_v:
        vid_next = path[j+1]
        edge = g.es.find(_between = ([vid], [vid_next]))
        dr = 'FROM' if edge.source==vid else 'TO'
        logging.debug(f"edge: {g.vs[edge.source]['doid']} {dr} {g.vs[edge.target]['doid']}")
      logging.debug(f"{j+1}. {v['doid']} ({v['name']}){dr}")
    vid_pa = PathRoot(g, path)
    logging.debug(f"path ancestor: {g.vs[vid_pa]['doid']}")
  return list(vs)

#############################################################################
def PathRoot(g, path):
  """Given a DAG path (NIDs), return NID for root, i.e. node with no target."""
  n_v=len(path)
  for j in range(n_v):
    vid=path[j]
    if j+1<n_v:
      vid_next = path[j+1]
      edge = g.es.find(_between = ([vid], [vid_next]))
      if edge.source==vid:
        return vid
    else:
      return vid
  return None

#############################################################################
def GetAncestors(g,vidxA):
  vA = g.vs[vidxA]
  vidxAncestors = {}
  for p in vA.neighbors(igraph.IN):
    vidxAncestors[p.index] = True
    vidxAncestors.update(GetAncestors(g,p.index))
  return vidxAncestors

#############################################################################
def ShowAncestry(g,vidxA,level):
  vA = g.vs[vidxA]
  logging.info(f"{level if level else '':>2} {'^'*level:>6} [{vA.index:>6}] {vA['id']:-14} ({vA['name']})")
  for p in vA.neighbors(igraph.IN):
    ShowAncestry(g, p.index, level+1)

#############################################################################
### GraphBase.bfs():
### Returns tuple:
###     The vertex IDs visited (in order)
###     The start indices of the layers in the vertex list
###     The parent of every vertex in the BFS
#############################################################################
#def BreadthFirstSearchTest(g):
#  rs = RootNodes(g)
#  if len(rs)>1:
#    logging.warning(f"multiple root nodes ({len(rs)}) using one only.")
#  r = rs[0];
#  bfs = g.bfs(r.index, mode=igraph.OUT)
#  vids, start_idxs, prnts = bfs
#  logging.debug(f"len(vids) = {len(vids)}; len(start_idxs) = {len(start_idxs)}; len(prnts) = {len(prnts)}')
#
#  logging.info(f"layers: {len(start_idxs)}")
#  start_idx_prev=0;
#  for layer,start_idx in enumerate(start_idxs):
#    logging.info(f"layer = {layer}")
#    for i in range(start_idx_prev,start_idx):
#      logging.info(f"\t{layer}) vs[{vids[i]}]: {g.vs[vids[i]]['id']} ({g.vs[vids[i]]['name']}); parent = {g.vs[prnts[vids[i]]]['id']} ({g.vs[prnts[vids[i]]]['name']})")
#    start_idx_prev=start_idx

#############################################################################
def InducedSubgraph(g, vs, implementation="auto"):
  for v in vs:
    logging.debug(f"{v['id']}: {v['name']}")
  logging.info(f"Selected nodes: {len(vs)}")
  g = g.induced_subgraph(vs, implementation=implementation)
  logging.info(f"SELECTED SUBGRAPH:  nodes: {g.vcount()}; edges: {g.ecount()}")
  return g

#############################################################################
def DisplayGraph(g,layout,w,h):
  """Layouts:
 "rt"          : reingold tilford tree
 "rt_circular" : reingold tilford circular
 "fr"          : fruchterman reingold
 "lgl"         : large_graph
"""
  visual_style = {}
  #color_dict = {"m": "blue", "f": "pink"}
  #for v in g.vs:
  #  v["gender"] = random.choice(('m', 'f'))
  #for e in g.es:
  #  e["is_formal"] = random.choice(range(0, 3))
  if layout=='kk':
    visual_style["layout"] = g.layout('kk')
  elif layout=='rt':
    visual_style["layout"] = g.layout('rt', 3) #tree depth?
  else:
    visual_style["layout"] = g.layout('large')

  visual_style["bbox"] = (w, h)
  visual_style["vertex_label"] = g.vs["name"]
  visual_style["vertex_size"] = 25
  #visual_style["vertex_size"] = [25+random.choice(range(-10, 10)) for v in g.vs]
  visual_style["vertex_color"] = "lightblue"
  #visual_style["vertex_color"] = [color_dict[gender] for gender in g.vs["gender"]]
  visual_style["vertex_shape"] = [random.choice(('rect', 'circle', 'rhombus')) for v in g.vs]
  #visual_style["edge_width"] = [1 + 2 * int(is_formal) for is_formal in g.es["is_formal"]]
  visual_style["edge_width"] = 2
  visual_style["margin"] = 20
  igraph.plot(g, **visual_style)

#############################################################################
def Graph2CyJsElements(g):
  """Convert igraph object to CytoscapeJS-compatible JSON "elements"."""
  def merge_dicts(x, y):
    return {**x, **y} #Python3
  nodes=[]; edges=[];
  for v in g.vs:
    nodes.append({'data':merge_dicts({'id':v['id']}, v.attributes())})
  for e in g.es:
    v_source = g.vs[e.source]
    v_target = g.vs[e.target]
    edges.append({'data':merge_dicts({'source':v_source['id'], 'target':v_target['id']}, e.attributes())})
  for node in nodes:
    if 'class' in node['data']:
      node['classes'] = node['data']['class']
  CyGraph = nodes + edges
  CyGraphJS = json.dumps(CyGraph, indent=2, sort_keys=False)
  return CyGraphJS

#############################################################################
def VisualStyle(g):
  for v in g.vs:
    v["gender"] = random.choice(('m', 'f'))
  for e in g.es:
    e["is_formal"] = random.choice(range(0, 3))
  visual_style = {}
  #visual_style["vertex_size"] = 25
  visual_style["vertex_size"] = [25+random.choice(range(-10, 10)) for v in g.vs]
  color_dict = {"m": "blue", "f": "pink"}
  visual_style["vertex_color"] = [color_dict[gender] for gender in g.vs["gender"]]
  visual_style["vertex_shape"] = [random.choice(('rect', 'circle', 'rhombus')) for v in g.vs]
  visual_style["vertex_label"] = g.vs["name"]
  visual_style["edge_width"] = [1 + 2 * int(is_formal) for is_formal in g.es["is_formal"]]
  visual_style["bbox"] = (700, 500)
  visual_style["margin"] = 20

  return visual_style

#############################################################################
def Plot(g, ofile_plot):
  vstyle = VisualStyle(g)
  if ofile_plot:
    igraph.plot(g, ofile_plot, **vstyle)
  else:
    igraph.plot(g, **vstyle) #interactive

#############################################################################
def Layout(g, method):
  if method=='kamada_kawai':
    visual_style["layout"] = g.layout("kk")
  elif method=='reingold_tilford':
    visual_style["layout"] = g.layout("rt",2)
  elif method=='reingold_tilford_circular':
    visual_style["layout"] = g.layout("rt_circular")
  elif method=='fruchterman_reingold':
    visual_style["layout"] = g.layout("fr")
  elif method=='large_graph':
    visual_style["layout"] = g.layout("lgl")

#############################################################################
def ComputeInfoContent(g):
  rs = RootNodes(g)
  if len(rs)>1:
    logging.warning(f"Multiple root nodes ({len(rs)}) using one only.")
  r = rs[0];
  ridx = r.index
  g.vs["ndes"] = [0 for i in range(len(g.vs))]
  g.vs["ic"] = [0.0 for i in range(len(g.vs))]
  NDescendants(g, ridx, 0)
  for v in g.vs:
    v["ic"] = -np.log10(min(float((v["ndes"]+1)/r["ndes"]), 1.0))
    v["ic"] = np.abs(v["ic"])

#############################################################################
def NDescendants(g, vidx, level):
  """Recursive by depth first search.  Since DAG may not be a tree, need
to avoid multiple counts via alternate parents."""
  dvidxs = set() #descendant vidxs
  for vidx_ in g.neighbors(vidx, mode=igraph.OUT):
    dvidxs.add(vidx_)
    dvidxs_this = NDescendants(g, vidx_, level+1)
    dvidxs |= dvidxs_this
    g.vs[vidx_]["ndes"] = len(dvidxs_this)
  ndes=len(dvidxs) #number of descendants of vidx
  logging.debug(f"{level*'>'}{level}) vs[{vidx}]: {g.vs[vidx]['id']} ({g.vs[vidx]['name']}); ndes = {ndes}")
  g.vs[vidx]["ndes"] = ndes
  return dvidxs

#############################################################################
def FindMICA(g, vidxA, vidxB, vidxFrom=None):
  """Start with root node as default MICA.  Self may be MICA.  If not, test children. 
Accumulate MICA list.  Recurse."""
  if vidxA==vidxB: return vidxA
  if not vidxFrom:
    r = RootNodes(g)[0] #should be only one
    vidxFrom = r.index
  vFrom = g.vs[vidxFrom] #provisional

  ##Short cut: If sole shared parent, is MICA:
  ##Also a kludge to avoid pathological recursion with DOID:265, DOID:268
  vidxA_parents = set(g.neighbors(vidxA, mode=igraph.IN))
  vidxB_parents = set(g.neighbors(vidxB, mode=igraph.IN))
  coparents = vidxA_parents & vidxB_parents
  if len(coparents)==1:
    vidxCop = list(coparents)[0]
    #logging.debug('returning coparent: %d'%(vidxCop))
    return vidxCop
  try:
    vidxAAncestors = GetAncestors(g, vidxA)
    vidxBAncestors = GetAncestors(g, vidxB)
  except Exception as e:
    logging.error(f"(Aack!): '{str(e)}'")
    raise
  if not (vidxFrom==vidxA or (vidxFrom in vidxAAncestors)):
    logging.error("(Aack!): vidxFrom not in vidxAAncestors.")
    logging.debug(f"vFrom: [{vidxFrom}] {vFrom['doid']} ({vFrom['name']})")
    logging.debug(f"vidxAAncestors: {str(vidxAAncestors)}")
    return None
  if not (vidxFrom==vidxB or (vidxFrom in vidxBAncestors)):
    logging.error("(Aack!): vidxFrom not in vidxBAncestors.")
    logging.debug(f"vFrom: [{vidxFrom}] {vFrom['doid']} ({vFrom['name']})")
    logging.debug(f"vidxBAncestors: {str(vidxBAncestors)}")
    return None
  if vidxA==vidxFrom or vidxB==vidxFrom: return vidxFrom
  micas = [] #list of tuples (vidx, ic)
  micas.append((vidxFrom, vFrom['ic']))
  vidxFrom_children = list(g.neighbors(vidxFrom, mode=igraph.OUT))
  for vidxFrom_child in vidxFrom_children:
    if (vidxFrom_child in vidxAAncestors) and (vidxFrom_child in vidxBAncestors):
      vidx_ = FindMICA(g, vidxA, vidxB, vidxFrom_child)
      if vidx_:
        v_ = g.vs[vidx_]
        micas.append((vidx_, v_['ic']))
  if not micas:
    logging.error('(Aack!): no MICAs found.')
    return None
  micas = sorted(micas, key=lambda x: -x[1]) #on 2nd field, descending
  return micas[0][0]

#############################################################################
def SimMatrixNodelist(g, fout):
  fout.write("vidx\tdoid\n")
  vidxs = [v.index for v in g.vs]
  vidxs.sort()
  for vidx in vidxs:
    v=g.vs[vidx]
    doid=v['doid']
    fout.write(f"{vidx}\t{doid}\n")
  logging.info(f"n_node: {len(vidxs)}")

#############################################################################
def SimMatrix(g, vidxA_query, skip, nmax, fout):
  """For every node-node pair in DAG, find MICA and write IC (similarity).
If vidxA specified, compute one row only."""
  fout.write("doidA\tdoidB\tdoidMICA\tsim\n")
  vidxs = [v.index for v in g.vs]
  vidxs.sort()
  n_in=0; n_out=0; n_nonzero=0; n_err=0;
  for i in range(len(vidxs)):
    if skip and i<skip: continue
    if nmax and (i-skip)==nmax: break
    vidxA = vidxs[i]
    if vidxA_query and vidxA_query!=vidxA: continue
    vA = g.vs[vidxA]
    doidA = vA['doid']
    logging.debug(f"vA: [{vidxA}] {vA['doid']} ({vA['name']})")
    n_nonzero_this=0
    n_in_this=0;
    for j in range(i+1, len(vidxs)):
      n_in+=1
      n_in_this+=1
      vidxB = vidxs[j]
      vB = g.vs[vidxB]
      doidB = vB['doid']
      logging.debug(f"vB: [{vidxB}] {vB['doid']} ({vB['name']})")
      try:
        vidxMICA = FindMICA(g, vidxA, vidxB, None)
      except Exception as e:
        logging.error(f"(Aack!): '{str(e)}'")
        vidxMICA = None
        pass
      if vidxMICA is None: #zero possible so use None
        logging.error(f"vA: [{vidxA}] {vA['doid']} ({vA['name']})")
        logging.error(f"vB: [{vidxB}] {vB['doid']} ({vB['name']})")
        n_err+=1
        continue
      vMICA = g.vs[vidxMICA]
      doidMICA = vMICA['doid']
      ic = vMICA['ic']
      if ic>0.0:
        n_nonzero_this+=1
        fout.write(f"{doidA.replace('DOID:','')}\t{doidB.replace('DOID:','')}\t{doidMICA.replace('DOID:','')}\t{ic:4f}\n")
        fout.flush()
        n_out+=1
      if (n_in%1e5)==0: logging.info(f"n_in: {n_in}; n_out: {n_out}; n_nonzero: {n_nonzero} ({100.0*n_nonzero/n_in:.1f}%%)")
    n_nonzero+=n_nonzero_this
    logging.debug(f"vA: [{vidxA}] {vA['doid']} ({vA['name']}); n_nonzero_this = {n_nonzero_this}/{n_in_this}; total n_nonzero = {n_nonzero}/{n_in} ({100.0*n_nonzero/n_in:.1f}%%)")
  logging.info(f"Total n_in: {n_in}; n_out: {n_out}; n_nonzero: {n_nonzero} ({100*n_nonzero/n_in:.1f}%%)")
  logging.info(f"Total n_err: {n_err}")



#!/usr/bin/env python3
"""
TIGA/TCRD db client utility 
"""
import os,sys,argparse,re,time,json,logging

from ...idg import tiga
from ...util import yaml as util_yaml
from ...util import db as util_db

#############################################################################
if __name__=='__main__':
  PARAM_FILE = os.environ['HOME']+"/.tcrd.yaml"
  epilog = "Example IDs: EFO_0004541, ENSG00000160785, ENSG00000215021"
  parser = argparse.ArgumentParser(description='TIGA/TCRD MySql client utility', epilog=epilog)
  ops = ['info',
	'listGenes',
	'listTraits',
	'getTraitAssociations',
	'getGeneAssociations',
	'getGeneTraitAssociations',
	'getGeneTraitProvenance',
	]
  parser.add_argument("op", choices=ops, help='OPERATION')
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--igene", dest="ifilegene", help="input gene ID file")
  parser.add_argument("--itrait", dest="ifiletrait", help="input trait ID file")
  parser.add_argument("--geneIds", help="input IDs, genes (ENSG)")
  parser.add_argument("--traitIds", help="input IDs, traits (EFO)")
  parser.add_argument("--param_file", default=PARAM_FILE, help=f"default param_file: {PARAM_FILE}")
  parser.add_argument("--dbhost")
  parser.add_argument("--dbport")
  parser.add_argument("--dbusr")
  parser.add_argument("--dbpw")
  parser.add_argument("--dbname")
  parser.add_argument("-v", "--verbose", dest="verbose", action="count", default=0)
  parser.add_argument("-q", "--quiet", action="store_true", help="Suppress progress notification.")
  args = parser.parse_args()

  # logging.PROGRESS = 15 (custom)
  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>0 else logging.ERROR if args.quiet else 15))

  params = util_yaml.ReadParamFile(args.param_file) if os.path.isfile(args.param_file) else {}
  if args.dbhost: params['DBHOST'] = args.dbhost 
  if args.dbport: params['DBPORT'] = args.dbport 
  if args.dbusr: params['DBUSR'] = args.dbusr 
  if args.dbpw: params['DBPW'] = args.dbpw
  if args.dbname: params['DBNAME'] = args.dbname 

  fout = open(args.ofile, "w+") if args.ofile else sys.stdout

  geneIds=None; traitIds=None;
  if args.ifilegene:
    fin = open(args.ifilegene)
    geneIds=[]
    while True:
      line = fin.readline()
      if not line: break
      geneIds.append(line.rstrip())
    logging.info(f"Input gene IDs: {len(geneIds)}")
    fin.close()
  elif args.geneIds:
    geneIds = re.split(r'[,\s]+', args.geneIds)

  if args.ifiletrait:
    fin = open(args.ifiletrait)
    traitIds=[]
    while True:
      line = fin.readline()
      if not line: break
      traitIds.append(line.rstrip())
    logging.info(f"Input trait IDs: {len(traitIds)}")
    fin.close()
  elif args.traitIds:
    traitIds = re.split(r'[,\s]+', args.traitIds)

#  try:
#    import mysql.connector as mysql
#    dbcon = mysql.connect(host=params['DBHOST'], port=params['DBPORT'], user=params['DBUSR'], passwd=params['DBPW'], db=params['DBNAME'])
#  except Exception as e:
#    logging.error(f'{e}')
#    try:
#      import MySQLdb as mysql
#      dbcon = mysql.connect(host=params['DBHOST'], port=int(params['DBPORT']), user=params['DBUSR'], passwd=params['DBPW'], db=params['DBNAME'])
#    except Exception as e2:
#      logging.error(f'{e2}')
#      sys.exit(1)

  try:
    dbcon = util_db.MySqlConnect(dbhost=params['DBHOST'], dbport=params['DBPORT'], dbusr=params['DBUSR'], dbpw=params['DBPW'], dbname=params['DBNAME'])
  except Exception as e:
    logging.error(f"Connect failed: {e}")
    sys.exit(1)

  if args.op=='info':
    tiga.Utils.Info(dbcon, fout)

  elif args.op=='listGenes':
    tiga.Utils.ListGenes(dbcon, fout)

  elif args.op=='listTraits':
    tiga.Utils.ListTraits(dbcon, fout)

  elif args.op=='getTraitAssociations':
    tiga.Utils.GetGeneTraitAssociations(None, traitIds, dbcon, fout)

  elif args.op=='getGeneAssociations':
    tiga.Utils.GetGeneTraitAssociations(geneIds, None, dbcon, fout)

  elif args.op=='getGeneTraitAssociations':
    tiga.Utils.GetGeneTraitAssociations(geneIds, traitIds, dbcon, fout)

  elif args.op=='getGeneTraitProvenance':
    tiga.Utils.GetGeneTraitProvenance(geneIds, traitIds, dbcon, fout)

  else:
    parser.error(f"Invalid operation: {args.op}")
    parser.print_help()



#!/usr/bin/env python3
"""
Utility functions for GlyCan REST API.
https://api.glygen.org/

curl -X GET "https://api.glygen.org/glycan/detail/G00053MO/" -H "accept: application/json"
"""
###
import sys,os,re,json,collections,time,urllib.parse,logging,tqdm,tqdm.auto
import pandas as pd
import requests
#
NCHUNK=100
#
API_HOST="api.glygen.org"
API_BASE_PATH=""
BASE_URL="https://"+API_HOST+API_BASE_PATH
#
##############################################################################
def GetGlycans(ids, skip, base_url=BASE_URL, fout=None):
  n_out=0; tags=None; df=None;
  for i in tqdm.auto.trange(len(ids), desc="IDs"):
    id_this = ids[i]
    response = requests.get(f"{base_url}/glycan/detail/{id_this}", headers={"Content-Type":"application/json"})
    result = response.json()
    logging.debug(json.dumps(result, indent=2))
    if not tags: tags = [tag for tag in result.keys() if type(result[tag]) not in (list, dict, collections.OrderedDict)]
    df_this = pd.DataFrame({tags[j]:[result[tags[j]]] for j in range(len(tags))})
    if fout is None: df = pd.concat([df, df_this])
    else: df_this.to_csv(fout, "\t", index=False, header=bool(n_out==0))
    n_out += df_this.shape[0]
  logging.info(f"n_out: {n_out}")
  return df

##############################################################################
def ListGlycans(skip, base_url=BASE_URL, fout=None):
  n_out=0; tags=None; df=None; tq=None;
  while True:
    url = f"""{base_url}/directsearch/glycan/?query={{"offset":{skip+1},"limit":{NCHUNK}}}"""
    response = requests.get(url, headers={"Content-Type":"application/json"})
    if response is None: break
    if response.status_code!=200:
      logging.debug(f"Status code: {response.status_code}")
      break
    results = response.json()
    logging.debug(json.dumps(results, indent=2))
    if len(results["results"])==0: break
    for result in results["results"]:
      if not tags:
        tags = list(result.keys())
        for tag in tags[:]:
          if type(result[tag]) in (list, dict, collections.OrderedDict):
            tags.remove(tag)
            logging.info(f"Ignoring tag: {tag}")
      if tq is None and "queryinfo" in results and "batch" in results["queryinfo"] and "total_hits" in results["queryinfo"]["batch"]:
        total_hits = results["queryinfo"]["batch"]["total_hits"]
        tq = tqdm.tqdm(total=(total_hits-skip))
      df_this = pd.DataFrame({tag:[result[tag] if tag in result else ""] for tag in tags})
      inchi_key = result["inchi_key"]["key"] if "inchi_key" in result and "key" in result["inchi_key"] else ""
      pubchem_cid=""; pubchem_sid=""; chebi_id=""; glytoucan_id="";
      if "crossref" in result:
        for crossref in result["crossref"]:
          if crossref["database"]=="ChEBI": chebi_id = crossref["id"]
          elif crossref["database"]=="PubChem Compound": pubchem_cid = crossref["id"]
          elif crossref["database"]=="PubChem Substance": pubchem_sid = crossref["id"]
          elif crossref["database"]=="GlyTouCan": glytoucan_id = crossref["id"]
      df_this = pd.concat([df_this,
	pd.DataFrame({"inchi_key":[inchi_key], "glytoucan_id":[glytoucan_id], "pubchem_cid":[pubchem_cid], "pubchem_sid":[pubchem_sid], "chebi_id":[chebi_id]})
	], axis=1)
      if fout is None: df = pd.concat([df, df_this])
      else: df_this.to_csv(fout, "\t", index=False, header=bool(n_out==0))
      n_out += df_this.shape[0]
      tq.update(n=df_this.shape[0])
    skip += NCHUNK
  tq.close()
  logging.info(f"n_out: {n_out}")
  return df

##############################################################################



#!/usr/bin/env python3
"""
CFChemDb utility functions.
"""
import os,sys,re,json,logging,yaml,tqdm
import pandas as pd
import psycopg2,psycopg2.extras

NCHUNK=100

#############################################################################
def Connect(dbhost, dbport, dbname, dbusr, dbpw):
  """Connect to db; specify default cursor type DictCursor."""
  dsn = (f"host='{dbhost}' port='{dbport}' dbname='{dbname}' user='{dbusr}' password='{dbpw}'")
  dbcon = psycopg2.connect(dsn)
  dbcon.cursor_factory = psycopg2.extras.DictCursor
  return dbcon

#############################################################################
def Version(dbcon, dbschema="public", fout=None):
  sql = (f"SELECT * FROM {dbschema}.dbversion")
  logging.debug(f"SQL: {sql}")
  df = pd.read_sql(sql, dbcon)
  if fout: df.to_csv(fout, "\t", index=False)
  return df

#############################################################################
def MetaListdbs(dbcon, fout=None):
  """Pg meta-command: list dbs from pg_database."""
  sql = ("SELECT pg_database.datname, pg_shdescription.description FROM pg_database LEFT OUTER JOIN pg_shdescription on pg_shdescription.objoid = pg_database.oid WHERE pg_database.datname ~ '^drug'")
  logging.debug(f"SQL: {sql}")
  df = pd.read_sql(sql, dbcon)
  if fout: df.to_csv(fout, "\t", index=False)
  return df

#############################################################################
def ListColumns(dbcon, dbschema="public", fout=None):
  df=None;
  sql1 = (f"SELECT table_name FROM information_schema.tables WHERE table_schema = '{dbschema}'")
  df1 = pd.read_sql(sql1, dbcon)
  for tname in df1.table_name:
    sql2 = (f"SELECT column_name,data_type FROM information_schema.columns WHERE table_schema = '{dbschema}' AND table_name = '{tname}'")
    df_this = pd.read_sql(sql2, dbcon)
    df_this["schema"] = dbschema
    df_this["table"] = tname
    df = df_this if df is None else pd.concat([df, df_this])
  df = df[["schema", "table", "column_name", "data_type"]]
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info(f"n_out: {df.shape[0]}")
  return df

#############################################################################
def ListTables(dbcon, dbschema="public", fout=None):
  '''Listing the tables.'''
  sql = (f"SELECT table_name FROM information_schema.tables WHERE table_schema = '{dbschema}'")
  df = pd.read_sql(sql, dbcon)
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info(f"n_out: {df.shape[0]}")
  return df

#############################################################################
def ListTablesRowCounts(dbcon, dbschema="public", fout=None):
  '''Listing the table rowcounts.'''
  df=None;
  sql1 = (f"SELECT table_name FROM information_schema.tables WHERE table_schema = '{dbschema}'")
  df1 = pd.read_sql(sql1, dbcon)
  for tname in df1.table_name:
    sql2 = (f"SELECT COUNT(*) AS rowcount FROM {dbschema}.{tname}")
    df_this = pd.read_sql(sql2, dbcon)
    df_this["schema"] = dbschema
    df_this["table"] = tname
    df = df_this if df is None else pd.concat([df, df_this])
  df = df[["schema", "table", "rowcount"]]
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info(f"n_out: {df.shape[0]}")
  return df

#############################################################################
def ListStructures(dbcon, dbschema="public", fout=None):
  sql = (f"SELECT id,name,cansmi FROM {dbschema}.mols")
  logging.debug(f"SQL: {sql}")
  df = pd.read_sql(sql, dbcon)
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info(f"n_out: {df.shape[0]}")
  return df

#############################################################################
def ListStructures2Smiles(dbcon, dbschema="public", fout=None):
  sql = (f"SELECT cansmi, id, name FROM {dbschema}.mols WHERE cansmi IS NOT NULL")
  logging.debug(f"SQL: {sql}")
  df = pd.read_sql(sql, dbcon)
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info(f"n_out: {df.shape[0]}")
  return df

#############################################################################
def GetStructure(dbcon, ids, fout=None):
  df=None;
  sql = ("""SELECT id,name,cansmi FROM mols WHERE id = '{}'""")
  for id_this in ids:
    logging.debug(sql.format(id_this))
    df_this = pd.read_sql(sql.format(id_this), dbcon)
    df = df_this if df is None else pd.concat([df, df_this])
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info(f"n_out: {df.shape[0]}")
  return df

#############################################################################



#!/usr/bin/env python3
"""
UMLS REST API client
UTS = UMLS Technology Services

https://documentation.uts.nlm.nih.gov/rest/home.html
https://documentation.uts.nlm.nih.gov/rest/authentication.html
https://documentation.uts.nlm.nih.gov/rest/concept/
https://documentation.uts.nlm.nih.gov/rest/source-asserted-identifiers/
https://documentation.uts.nlm.nih.gov/rest/search/
https://www.nlm.nih.gov/research/umls/knowledge_sources/metathesaurus/release/abbreviations.html

TGT = Ticket Granting Ticket
(API requires one ticket per request.)
CUI = Concept Unique Identifier
Atom is a term in a source.
Term-to-concept is many to one.

Retrieves information for a known Semantic Type identifier (TUI)
/semantic-network/{version}/TUI/{id}
(DOES NOT SEARCH FOR INSTANCES OF THIS TYPE -- RETURNS METADATA ONLY.)
Example TUIs:
CHEM|Chemicals & Drugs|T116|Amino Acid, Peptide, or Protein
CHEM|Chemicals & Drugs|T195|Antibiotic
CHEM|Chemicals & Drugs|T123|Biologically Active Substance
CHEM|Chemicals & Drugs|T122|Biomedical or Dental Material
CHEM|Chemicals & Drugs|T103|Chemical
CHEM|Chemicals & Drugs|T120|Chemical Viewed Functionally
CHEM|Chemicals & Drugs|T104|Chemical Viewed Structurally
CHEM|Chemicals & Drugs|T200|Clinical Drug
CHEM|Chemicals & Drugs|T126|Enzyme
CHEM|Chemicals & Drugs|T125|Hormone
CHEM|Chemicals & Drugs|T129|Immunologic Factor
CHEM|Chemicals & Drugs|T130|Indicator, Reagent, or Diagnostic Aid
CHEM|Chemicals & Drugs|T114|Nucleic Acid, Nucleoside, or Nucleotide
CHEM|Chemicals & Drugs|T109|Organic Chemical
CHEM|Chemicals & Drugs|T121|Pharmacologic Substance
CHEM|Chemicals & Drugs|T192|Receptor
CHEM|Chemicals & Drugs|T127|Vitamin
DISO|Disorders|T020|Acquired Abnormality
DISO|Disorders|T190|Anatomical Abnormality
DISO|Disorders|T049|Cell or Molecular Dysfunction
DISO|Disorders|T019|Congenital Abnormality
DISO|Disorders|T047|Disease or Syndrome
DISO|Disorders|T050|Experimental Model of Disease
DISO|Disorders|T033|Finding
DISO|Disorders|T037|Injury or Poisoning
DISO|Disorders|T048|Mental or Behavioral Dysfunction
DISO|Disorders|T191|Neoplastic Process
DISO|Disorders|T046|Pathologic Function
DISO|Disorders|T184|Sign or Symptom
GENE|Genes & Molecular Sequences|T087|Amino Acid Sequence
GENE|Genes & Molecular Sequences|T088|Carbohydrate Sequence
GENE|Genes & Molecular Sequences|T028|Gene or Genome
GENE|Genes & Molecular Sequences|T085|Molecular Sequence
GENE|Genes & Molecular Sequences|T086|Nucleotide Sequence

https://github.com/HHS/uts-rest-api
https://utslogin.nlm.nih.gov

https://www.nlm.nih.gov/research/umls/knowledge_sources/metathesaurus/release/abbreviations.html
Some term types:

 CE : Entry term for a Supplementary Concept
 ET : Entry term
 FN : Full form of descriptor
 HG : High Level Group Term
 HT : Hierarchical term
 LLT : Lower Level Term
 MH : Main heading
 MTH_FN : MTH Full form of descriptor
 MTH_HG : MTH High Level Group Term
 MTH_HT : MTH Hierarchical term
 MTH_LLT : MTH Lower Level Term
 MTH_OS : MTH System-organ class
 MTH_PT : Metathesaurus preferred term
 MTH_SY : MTH Designated synonym
 NM : Name of Supplementary Concept
 OS : System-organ class
 PCE : Preferred entry term for Supplementary Concept
 PEP : Preferred entry term
 PM : Machine permutation
 PT : Designated preferred name
 PTGB : British preferred term
 SY : Designated synonym
 SYGB : British synonym
 
Some relationships:

 RB : has a broader relationship
 RL : the relationship is similar or "alike". 
 RN : has a narrower relationship
 RO : has relationship other than synonymous, narrower, or broader
 RQ : related and possibly synonymous.
 RU : Related, unspecified
 SY : source asserted synonymy.
"""
###
import sys,os,re,yaml,json,urllib.parse,csv,logging,requests,time
import pandas as pd
from functools import total_ordering
#
from lxml import etree
from pyquery import PyQuery
#
from ..util import rest
#
###
API_HOST='uts-ws.nlm.nih.gov'
API_BASE_PATH="/rest"
API_BASE_URL=f"https://{API_HOST}{API_BASE_PATH}"
API_AUTH_SERVICE="http://umlsks.nlm.nih.gov"
API_AUTH_HOST="utslogin.nlm.nih.gov"
API_AUTH_ENDPOINT='/cas/v1/api-key'
API_VERSION="current"
API_HEADERS={"Content-type": "application/x-www-form-urlencoded", "Accept": "text/plain", "User-Agent":"python" }
#
## elements common to 'Concept' and 'SourceAtomCluster' class
UMLS_COMMON_FIELDS=['classType','name','ui','atomCount','definitions','atoms','defaultPreferredAtom']
UMLS_OPTIONAL_FIELDS=['parents','children','relations','descendants']
#
#############################################################################
@total_ordering
class Atom:
  def __init__(self, cui, src, code, name):
    self.cui = cui
    self.src = src
    self.code = code
    self.name = name

  def __eq__(self, other):
    return ((self.cui, self.src, self.code) == (other.cui, other.src, other.code))

  def __ne__(self, other):
    return not (self == other)

  def __lt__(self, other):
    return ((self.cui, self.src, self.code) < (other.cui, other.src, other.code))
  def __hash__(self):
    return id(self)

#############################################################################
class SourceList:
  def __init__(self):
    self.sources=[];

  def initFromFile(self, sfile):
    fin = open(sfile)
    if not fin:
      logging.error(f'Could not open {sfile}')
      return
    csvReader = csv.reader(fin, delimiter='\t', quotechar='"')
    row = csvReader.next() #ignore header
    while True:
      try:
        row = csvReader.next()
      except:
        break
      self.sources.append(tuple(row))
    self.sources.sort()
    fin.close()

  def initFromApi(self, base_url, ver, auth):
    url = (f'{base_url}/metadata/{ver}/sources')
    tgt = auth.gettgt()
    response = UmlsApiGet(url, auth, tgt)
    response.encoding = 'utf-8'
    items = json.loads(response.text)
    logging.debug(json.dumps(items, indent=4))
    sources = items["result"]
    tags = ["abbreviation", "shortName", "preferredName"]
    for source in sources:
      row=[]
      for tag in tags:
        row.append(source[tag] if tag in source else '')
      self.sources.append(tuple(row))
    self.sources.sort()

  def has_src(self,_src):
    for abbr,name,ver in self.sources:
      if _src == abbr:
        return True
    return False

#############################################################################
class Authentication:
  def __init__(self, apikey, service, url, headers):
    self.apikey=apikey
    self.service=service
    self.url=url
    self.headers=headers
    self.verbosity=0

  def gettgt(self):
    response = requests.post(self.url, data={'apikey':self.apikey}, headers=self.headers)
    logging.debug(f'response = {response}')
    d = PyQuery(response.text)

    if response.status_code not in (200, 201):
      logging.error(f'HTTP RESPONSE CODE: {response.status_code} ({d})')
      return None
      
    ## Extract the entire URL needed (action attribute),
    ## to make a POST call to this URL in the getst method.
    tgt = d.find('form').attr('action')
    logging.debug(f'tgt = {tgt}')
    return tgt

  def getst(self, tgt):
    r = requests.post(tgt, data={'service':self.service}, headers=self.headers)
    st = r.text
    return st

  def setVerbosity(self, v):
    self.verbosity=v

#############################################################################
def UmlsAuthGetTicket(auth, tgt, tries=10, sleep=1):
  for i in range(1, tries+1):
    try:
      tkt = auth.getst(tgt)
      return tkt
    except Exception as e:
      logging.error(f'{i}. {e}')
      time.sleep(sleep)
      continue
  return None

#############################################################################
def UmlsApiGet(url, auth, tgt, params={}, tries=10, sleep=1):
  for i in range(1,tries+1):
    try:
      params['ticket'] = UmlsAuthGetTicket(auth, tgt)
      response = requests.get(url, params=params)
      return response
    except Exception as e:
      logging.error(f'{i}. {e}')
      time.sleep(sleep)
      continue
  return None

#############################################################################
def ReadParamFile(fparam):
  params={};
  with open(fparam, 'r') as fh:
    for param in yaml.load_all(fh, Loader=yaml.BaseLoader):
      for k,v in param.items():
        params[k] = v
  return params

#############################################################################
def XrefConcept(src, ids, skip, nmax, auth, ver=API_VERSION, base_url=API_BASE_URL, fout=None):
  """Find UMLS concept/CUI from external source cross-reference."""
  n_in=0; n_concept=0; n_out=0;
  result_tags=None;
  tgt = auth.gettgt()
  for id_query in ids:
    n_in+=1
    if nmax and n_in>nmax: break
    if skip and n_in<=skip:
      logging.debug(f'[{id_query}] skipping...')
      continue
    url = (f"{base_url}/content/{ver}/source/{src}/{id_query}" if src else f"/CUI/{id_query}")
    logging.debug(f'{n_in}. url="{url}"')
    response = UmlsApiGet(url, auth, tgt)
    response.encoding = 'utf-8'
    try:
      items = json.loads(response.text)
    except Exception as e:
      logging.info(f'{n_in}. [{id_query}] {e}')
      logging.debug(f'response.text="{response.text}"')
      continue
    logging.debug(json.dumps(items, indent=4))
    result = items["result"]
    for key in UMLS_COMMON_FIELDS+UMLS_OPTIONAL_FIELDS:
      logging.info(f"""{key:14s}: {(result[key] if key in result else '')}""") 
    if 'semanticTypes' in result:
      for i,styp in enumerate(result["semanticTypes"]):
        for key in styp.keys():
          logging.info(f'Semantic type {i+1}. {key}: {styp[key]}')
    if n_out==0 or not result_tags:
      result_tags = result.keys()
      id_tag = (f'{src}_id' if src else 'CUI')
      fout.write('\t'.join([id_tag]+result_tags)+'\n')
    vals = [id_query]
    for tag in result_tags:
      val = (result[tag] if tag in result else '')
      if tag == 'concepts':
        url2 = result['concepts']
        response2 = UmlsApiGet(url2, auth, tgt)
        response2.encoding = 'utf-8'
        items2  = json.loads(response2.text)
        logging.debug(json.dumps(items2, indent=4))
        result2 = items2['result']
        concepts = result2['results'] if 'results' in result2 else []
        cuis=[]
        for concept in concepts:
          cui = concept['ui'] if 'ui' in concept else None
          if cui: n_concept+=1
          cuis.append(cui)
        val=';'.join(cuis)
      elif tag=='semanticTypes':
        if type(val) is list:
          sts=[]
          for st in val:
            if type(st) is dict and 'name' in st: sts.append(st['name'])
          val = ';'.join(sts)
      elif tag in ('relations', 'parents', 'children','descendants','ancestors',
	'attributes', 'contentViewMemberships', 'atoms', 'defaultPreferredAtom',
	'definitions'):
        if val != 'NONE': val = ''
      else:
        if type(val) is str:
          val = val.replace(base_url, '')
      vals.append(str(val))
    fout.write('\t'.join(vals)+'\n')
    n_out+=1
  logging.info(f'n_concept: {n_concept}')
  logging.info(f'n_out: {n_out}')

#############################################################################
def GetCodes(cuis, srcs, auth, ver=API_VERSION, base_url=API_BASE_URL, fout=None):
  n_out=0; df=None;
  for cui in cuis:
    codes = Cui2Code(cui, srcs, auth, ver, base_url)
    for src in sorted(codes.keys()):
      atoms = sorted(list(codes[src]))
      for atom in atoms:
        df_this = pd.DataFrame({'CUI':[cui], 'src':[src], 'atom_code':atom.code, 'atom_name':atom.name})
        if fout is None: df = pd.concat([df, df_this])
        else: df_this.to_csv(fout, "\t", index=False, header=bool(n_out==0))
        n_out+=1
  logging.info(f'n_cui: {len(cuis)}')
  logging.info(f'n_out: {n_out}')
  if fout is None: return df

#############################################################################
def Cui2Code(cui, srcs, auth, ver=API_VERSION, base_url=API_BASE_URL):
  n_atom=0; pNum=1; params={}; codes={};
  if srcs: params['sabs']=srcs
  url = (f'{base_url}/content/{ver}/CUI/{cui}/atoms')
  tgt = auth.gettgt()
  while True:
    params['pageNumber'] = pNum
    response = UmlsApiGet(url, auth, tgt, params=params)
    response.encoding = 'utf-8'
    try:
      items = json.loads(response.text)
    except Exception as e:
      logging.error(str(e))
      break
    logging.debug (json.dumps(items, indent=4))
    atoms = items["result"] if "result" in items else []
    for atom in atoms:
      n_atom+=1
      src = atom['rootSource'] if 'rootSource' in atom else None
      code = atom['code'] if 'code' in atom else None
      if code: code = re.sub(r'^.*/', '', code)
      name = atom['name'] if 'name' in atom else None
      cui = atom['concept'] if 'concept' in atom else None
      if cui: cui = re.sub(r'^.*/', '', cui)
      a = Atom(cui, src, code, name)
      if not src in codes: codes[src] = set()
      codes[src].add(a)
    pageSize = items["pageSize"] if "pageSize" in items else None
    pageNumber = items["pageNumber"] if "pageNumber" in items else None
    pageCount = items["pageCount"] if "pageCount" in items else None
    if pageNumber!=pNum:
      logging.error(f'pageNumber!=pNum ({pageNumber}!={pNum})')
      break
    elif pNum==pageCount:
      logging.debug(f'(done) pageNumber==pageCount ({pageNumber})')
      break
    else:
      pNum+=1
  return codes

#############################################################################
def GetAtoms(cuis, skip, nmax, srcs, auth, ver=API_VERSION, base_url=API_BASE_URL, fout=None):
  '''Expected fields: sourceDescriptor,suppressible,code,name,language,descendants,classType,sourceConcept,obsolete,relations,parents,children,concept,ui,rootSource,definitions,attributes,ancestors,termType'''
  n_in=0; n_atom=0; n_out=0; df=None; tags=None;
  tgt = auth.gettgt()
  for cui in cuis:
    n_in+=1
    if skip and n_in<=skip:
      logging.debug(f'[{cui}] skipping...')
      continue
    pNum=1;
    url = (f'{base_url}/content/{ver}/CUI/{cui}/atoms')
    params={}
    if srcs: params['sabs']=srcs
    while True:
      params['pageNumber']=pNum
      response = UmlsApiGet(url, auth, tgt, params=params)
      if not response:
        break
      if response.status_code != 200:
        logging.error(f'response.status_code = "{response.status_code}"')
        break
      response.encoding = 'utf-8'
      logging.debug(f'params = {str(params)}')
      logging.debug(response.text)
      items = json.loads(response.text)
      logging.debug(json.dumps(items, indent=4))
      result = items["result"]
      for atom in result:
        n_atom+=1
        if not tags: tags = list(atom.keys())
        for tag in ('relations', 'parents', 'children','descendants','ancestors', 'attributes', 'contentViewMemberships'):
            if tag in atom and atom[tag]!='NONE': atom[tag] = '*'
        df_this = pd.DataFrame({tag:[atom[tag] if tag in atom else ''] for tag in tags})
        if fout is None: df = pd.concat([df, df_this])
        else: df_this.to_csv(fout, "\t", index=False, header=bool(n_out==0))
        n_out+=1
      pageSize = items["pageSize"]
      pageNumber = items["pageNumber"]
      pageCount = items["pageCount"]
      if pageNumber!=pNum:
        logging.error(f'pageNumber!=pNum ({pageNumber}!={pNum})')
        break
      elif pNum==pageCount:
        logging.debug(f'(done) pageNumber==pageCount ({pageNumber})')
        break
      else:
        pNum+=1
    if nmax and n_in==nmax: break
  logging.info(f'n_atom: {n_atom}')
  logging.info(f'n_out: {n_out}')
  if fout is None: return df

#############################################################################
def GetRelations(cuis, skip, nmax, srcs, auth, ver=API_VERSION, base_url=API_BASE_URL, fout=None):
  n_in=0; n_rel=0; n_out=0; tags=None; df=None;
  tgt = auth.gettgt()
  for cui in cuis:
    n_in+=1
    if skip and n_in<=skip:
      logging.debug(f'[{cui}] skipping...')
      continue
    pNum=1;
    url = (f'{base_url}/content/{ver}/CUI/{cui}/relations')
    params={}
    if srcs: params['sabs']=srcs
    while True:
      params['pageNumber']=pNum
      response = UmlsApiGet(url, auth, tgt, params=params)
      if not response:
        break
      if response.status_code != 200:
        logging.error(f'response.status_code = "{response.status_code}"')
        break
      response.encoding = 'utf-8'
      logging.debug(f'params = {str(params)}')
      logging.debug(f'{response.text}')
      items = json.loads(response.text)
      logging.debug(json.dumps(items, indent=4))
      result = items["result"]
      for rel in result:
        n_rel+=1
        if not tags: tags = list(rel.keys())
        df_this = pd.DataFrame({tag:[rel[tag] if tag in rel else ''] for tag in tags})
        df_this["cui"] = [cui]
        if fout is None: df = pd.concat([df, df_this])
        else: df_this.to_csv(fout, "\t", index=False, header=bool(n_out==0))
        n_out+=1
      pageSize = items["pageSize"]
      pageNumber = items["pageNumber"]
      pageCount = items["pageCount"]
      if pageNumber!=pNum:
        logging.error(f': pageNumber!=pNum ({pageNumber}!={pNum})')
        break
      elif pNum==pageCount:
        break
      else:
        pNum+=1
    if nmax and n_in==nmax: break
  logging.info(f'n_rel: {n_rel}')
  logging.info(f'n_out: {n_out}')
  if fout is None: return df

#############################################################################
def Search(query, searchType, inputType, returnIdType, srcs, auth, ver=API_VERSION, base_url=API_BASE_URL, fout=None):
  """Retrieves CUIs for a search term.
Expected fields: ui, rootSource, name, uri.
See https://documentation.uts.nlm.nih.gov/rest/search/
"""
  src_counts={}; n_item=0; n_out=0; pNum=1; tags=None; df=None;
  url = (f'{base_url}/search/{ver}')
  params = {'string':query, 'searchType':searchType, 'inputType':inputType, 'returnIdType':returnIdType}
  if srcs: params['sabs'] = srcs
  tgt = auth.gettgt()
  while True:
    params['pageNumber']=pNum
    logging.debug(f'params = {str(params)}')
    response = UmlsApiGet(url, auth, tgt, params=params)
    response.encoding = 'utf-8'
    items = json.loads(response.text)
    logging.debug (json.dumps(items, indent=4))
    result = items['result']
    classType = result['classType']
    pageSize = items["pageSize"]
    pageNumber = items["pageNumber"]
    ##No pageCount in search response.
    items = result['results']
    if not items:
      break
    elif len(items)==1 and items[0]['name']=='NO RESULTS':
      break
    elif pageNumber!=pNum:
      logging.debug(f'pageNumber!=pNum ({pageNumber}!={pNum})')
      break
    for item in items:
      n_item+=1
      if not tags: tags = list(item.keys())
      vals = []
      cui = item['ui'] if 'ui' in item else None
      if 'rootSource' in item:
        if not item['rootSource'] in src_counts:
          src_counts[item['rootSource']]=0
        src_counts[item['rootSource']]+=1
      df_this = pd.DataFrame({tag:[item[tag] if tag in item else ''] for tag in tags})
      if fout is None: df = pd.concat([df, df_this])
      else: df_this.to_csv(fout, "\t", index=False, header=bool(n_out==0))
      n_out+=1
    pNum+=1
  logging.info(f'n_item: {n_item}')
  logging.info(f'n_out: {n_out}')
  for src in sorted(src_counts.keys()):
    logging.info(f'{src:16s}: {src_counts[src]:3d} items')
  if fout is None: return df

#############################################################################



#!/usr/bin/env python3
"""
http://igraph.org/python/doc/tutorial/tutorial.html

Note that igraph defines its own IDs.  These are integers, not the same
as imported GraphML "id" values.

Compute information content (IC) for nodes in a directed acyclic graph
(DAG).
For each node, information content (IC) depends on total number of
descendants ndes, +1 for self.  Then IC is -log10((ndes+1)/n_total).
Given this we can identify the most informative common ancestor
(MICA), and the semantic similarity between any two nodes/terms.
Note that this is a kind of "Platonic" IC; in contrast, a set of
annotated disease instances could have a sampled IC which depends
on the frequency of each DO annotation.  In that scenario, an unused
disease has no effect on the sampled IC.

Used for Disease Ontology, as converted to GraphML with
BioClients.util.obo.App and Go_do_graph_IC.sh.
"""
#############################################################################
import sys,os,argparse,logging,re

from .. import igraph as util_igraph

#############################################################################
def TestMICA(g, nidA, nidB):
  vA = g.vs.find(id = nidA)
  vB = g.vs.find(id = nidB)
  logging.info(f"\tvA: [{vA.index}] {vA['doid']} ({vA['name']})")
  logging.info(f"\tvB: [{vB.index}] {vB['doid']} ({vB['name']})")
  vidx_mica = util_igraph.FindMICA(g, vA.index, vB.index, None)
  v = g.vs[vidx_mica]
  logging.info(f"MICA: [{vidx_mica}] {v['doid']} ({v['name']}); IC = {v['ic']:.3f}")

#############################################################################
if __name__=='__main__':
  epilog="""\
OPERATIONS:
summary: summary of graph;
degree_distribution: degree distribution;
node_select: select for nodes by criteria;
edge_select: select for edges by criteria;
connectednodes: connected node[s];
disconnectednodes: disconnected node[s];
rootnodes: root node[s] of DAG;
topnodes: root node[s] & children of DAG;
shortest_paths: shortest paths, nodes A ~ B;
show_ancestry: show ancestry, node A;
graph2cyjs: CytoscapeJS JSON;
NOTE: select also deletes non-matching for modified output.
Info content (IC) and most informative common ancestor (MICA) for directed
acyclic graph (DAG).
simMatrixNodelist outputs vertex indices with node IDs.  simMatrix with --nidA to compute one row.
"""
  parser = argparse.ArgumentParser(description="IGraph (python-igraph API) utility, graph processingand display", epilog=epilog)
  ops = ["summary",
	"degree_distribution",
	"rootnodes",
	"topnodes",
  	"graph2cyjs",
	"shortest_paths",
	"show_ancestry",
	"connectednodes",
	"disconnectednodes",
	"node_select",
	"edge_select",
	"ic_computeIC", "ic_findMICA", "ic_simMatrix", "ic_simMatrixNodelist", "ic_test" ]
  parser.add_argument("op", choices=ops, help='OPERATION')
  parser.add_argument("--i", dest="ifile", required=True, help="input file or URL (e.g. GraphML)")
  parser.add_argument("--o", dest="ofile", help="output file")
  parser.add_argument("--selectfield", help="field (attribute) to select")
  parser.add_argument("--selectquery", help="string query")
  parser.add_argument("--selectval", type=float, help="numerical query")
  parser.add_argument("--select_exact", help="exact string match (else substring/regex)")
  parser.add_argument("--select_equal", action="store_true", help="numerical equality select")
  parser.add_argument("--select_lt", action="store_true", help="numerical less-than select")
  parser.add_argument("--select_gt", action="store_true", help="numerical greater-than select")
  parser.add_argument("--select_negate", action="store_true", help="negate select criteria")
  parser.add_argument("--display", action="store_true", help="display graph interactively")
  parser.add_argument("--display_width", type=int, default=700)
  parser.add_argument("--display_height", type=int, default=500)
  parser.add_argument("--display_layout", default="rt")
  parser.add_argument("--depth", type=int, default=1, help="depth for --topnodes")
  parser.add_argument("--nidA", help="nodeA ID")
  parser.add_argument("--nidB", help="nodeB ID")
  parser.add_argument("--nmax", type=int)
  parser.add_argument("--skip", type=int)
  parser.add_argument("--recursionlimit", type=int, default=sys.getrecursionlimit())
  parser.add_argument("--quiet", action="store_true")
  parser.add_argument("-v", "--verbose", dest="verbose", action="count", default=0)
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>0 else logging.ERROR if args.quiet else 15))

  fin = open(args.ifile)
  fout = open(args.ofile, "w+") if args.ofile else sys.stdout

  sys.setrecursionlimit(args.recursionlimit)

  ###
  #Load graph:
  g = util_igraph.Load_GraphML(args.ifile)
  vs=[]; #vertices for subgraph selection

  if args.op=='summary':
    util_igraph.GraphSummary(g)

  elif args.op=='degree_distribution':
    util_igraph.DegreeDistribution(g)

  elif args.op=='node_select':
    if args.selectquery:
      vs = util_igraph.NodeSelect_String(g, args.selectfield, args.selectquery, args.select_exact, args.select_negate)
    elif args.selectval:
      parser.error('Numerical select not implemented yet.')
    else:
      parser.error('Select query or value required.')
    g = util_igraph.InducedSubgraph(g, vs, implementation="auto")
    util_igraph.Save_GraphML(g, fout)

  elif args.op=='edge_select':
    parser.error(f"Operation not implemented yet: {args.op}")

  elif args.op=='connectednodes':
    vs = util_igraph.ConnectedNodes(g)
    g = util_igraph.InducedSubgraph(g, vs, implementation="auto")
    util_igraph.Save_GraphML(g, fout)

  elif args.op=='disconnectednodes':
    vs = util_igraph.DisconnectedNodes(g)
    g = util_igraph.InducedSubgraph(g, vs, implementation="auto")
    util_igraph.Save_GraphML(g, fout)

  elif args.op=='rootnodes':
    vs = util_igraph.RootNodes(g)
    g = util_igraph.InducedSubgraph(g, vs, implementation="auto")
    util_igraph.Save_GraphML(g, fout)

  elif args.op=='topnodes':
    vs = util_igraph.TopNodes(g, args.depth)
    g = util_igraph.InducedSubgraph(g, vs, implementation="auto")
    util_igraph.Save_GraphML(g, fout)

  elif args.op=='shortest_paths':
    if not (args.nidA and  args.nidB): parser.error('--shortest_path requires nidA and nidB.')
    vs = util_igraph.ShortestPaths(g, args.nidA, args.nidB)
    g = util_igraph.InducedSubgraph(g, vs, implementation="auto")
    util_igraph.Save_GraphML(g, fout)

  elif args.op=='show_ancestry':
    if not nidA: parser.error('--show_ancestry requires nidA.')
    vA = g.vs.find(id=args.nidA)
    vidxA = vA.index
    util_igraph.ShowAncestry(g, vidxA, 0)

  elif args.op=='graph2cyjs':
    fout.write(util_igraph.Graph2CyJsElements(g))

  elif args.op == 'ic_computeIC':
    if not g.is_dag(): parser.error(f"Graph not DAG; required for operation: {args.op}")
    util_igraph.ComputeInfoContent(g)
    util_igraph.Save_GraphML(g, fout)

  elif args.op == 'ic_findMICA':
    if not g.is_dag(): parser.error(f"Graph not DAG; required for operation: {args.op}")
    if not (args.nidA and args.nidB):
      parser.error('findMICA requires --nidA and --nidB.')
      parser.print_help()
    vA = g.vs.find(id = args.nidA)
    vB = g.vs.find(id = args.nidB)
    logging.debug(f"\tvA: [{vA.index}] {vA['doid']} ({vA['name']})")
    logging.debug(f"\tvB: [{vB.index}] {vB['doid']} ({vB['name']})")
    vidx_mica = util_igraph.FindMICA(g, vA.index, vB.index, None)
    v = g.vs[vidx_mica]
    logging.info(f"MICA: [{vidx_mica}] {v['doid']} ({v['name']}); IC = {v['ic']:.3f}")

  elif args.op == 'ic_simMatrix':
    vidxA = g.vs.find(id=args.nidA).index if args.nidA else None
    util_igraph.SimMatrix(g, vidxA, args.skip, args.nmax, fout)

  elif args.op == 'ic_simMatrixNodelist':
    util_igraph.SimMatrixNodelist(g, fout)

  elif args.op == 'test':
    #if not (args.nidA and args.nidB):
    #  parser.error('test requires --nidA and --nidB.')
    #  parser.print_help()
    import cProfile
    #cProfile.run('TestMICA(g,"%s","%s")'%(args.nidA,args.nidB))
    cProfile.run('SimMatrix(g, 0, 1, fout)')
    #cProfile.runctx('TestMICA(g,"%s","%s")'%(args.nidA,args.nidB), globals(), locals())

  else:
    parser.error('No operation specified.')
    parser.print_help()

  fout.close()

  if args.display:
    util_igraph.DisplayGraph(g, args.display_layout, args.display_width, args.display_height)




#!/usr/bin/env python3
"""
Pandas Csv2Html.
"""
###
import sys,os,argparse,re,logging
import pandas as pd

#############################################################################
def CSS():
  return """\
/* includes alternating gray and white with on-hover color */

.mystyle {
  font-size: 11pt; 
  font-family: Arial;
  border-collapse: collapse; 
  border: 1px solid silver;
}

.mystyle table {
  width: 100%;
}

.mystyle td, th {
  padding: 5px;
  max-width:400px;
  word-wrap:break-word;
  overflow-wrap: break-word;
}

.mystyle tr:nth-child(even) {
  background: #E0E0E0;
}

.mystyle tr:hover {
  background: silver;
  cursor: pointer;
}
"""

# default=lambda(x: f"{x:.3f}"),
#############################################################################
if __name__=='__main__':
  parser = argparse.ArgumentParser(description='Pandas CSV2HTML.')
  parser.add_argument("--i", dest="ifile", required=True, help="input (CSV|TSV)")
  parser.add_argument("--o", dest="ofile", help="output (HTML)")
  parser.add_argument("--csv", action="store_true", help="delimiter is comma")
  parser.add_argument("--tsv", action="store_true", help="delimiter is tab")
  parser.add_argument("--title", help="HTML title")
  parser.add_argument("--nrows", type=int)
  parser.add_argument("--skiprows", type=int)
  parser.add_argument("--justify", choices=["center","right","left"], default="center")
  parser.add_argument("--classes", default="mystyle", help="CSS classes")
  parser.add_argument("--columns", help="Subset of columns to write (comma delimited)")
  parser.add_argument("--index_columns", help="Index columns for hierarchical (comma delimited)")
  parser.add_argument("--na_rep", default="", help="String representation of NaN")
  parser.add_argument("--float_format", help="Function(float) -> string.")
  parser.add_argument("--border", type=int, default=1, help="Border size.")
  parser.add_argument("--table_id", help="CSS ID.")
  parser.add_argument("--sparsify", action="store_true", help="Set to False for a DataFrame with a hierarchical index to print every multiindex key at each row.")
  parser.add_argument("-v", "--verbose", action="count", default=0)
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  fout = open(args.ofile, "w") if args.ofile else sys.stdout

  if args.csv: delim=','
  elif args.tsv: delim='\t'
  elif re.search('\.csv', args.ifile, re.I): delim=','
  else: delim='\t'

  title = args.title if args.title else f"Csv2Html: {os.path.basename(args.ifile)}"
  columns = re.split(r',', args.columns) if args.columns else None

  df = pd.read_csv(args.ifile, sep=delim, nrows=args.nrows, skiprows=args.skiprows)

  if args.index_columns is not None:
    df.set_index(re.split(r',', args.index_columns), drop=True, append=False, inplace=True)

  table_html = df.to_html(header=True,
	index=(args.index_columns is not None), 
	formatters=None, float_format=args.float_format, na_rep=args.na_rep,
	columns=columns, sparsify=args.sparsify,
	bold_rows=True, border=args.border, justify=args.justify,
	classes=args.classes, table_id=args.table_id, render_links=True,
	show_dimensions=False)

  pd.set_option('colheader_justify', 'center')   # FOR TABLE <th>

  html = f"""
<!DOCTYPE html>
<html>
  <head><title>{title}</title></head>
  <style type="text/css">{CSS()}</style>
  <body>
    {table_html}
  </body>
</html>
"""

  fout.write(html)
  fout.close()




#!/usr/bin/env python3
"""
Monarch BioLink REST API client
https://api.monarchinitiative.org/api/
https://github.com/monarch-initiative/biolink-api/
"""
###
import sys,os,argparse,json,csv,logging
#
from ..util import rest
#
API_HOST='api.monarchinitiative.org'
API_BASE_PATH='/api'
#
#############################################################################
def GetDisease(base_url, ids, args, fout):
  csvWriter = csv.writer(fout, delimiter=',', quotechar='"', quoting=csv.QUOTE_NONNUMERIC)
  tags=None;
  i_ent=0; n_assn=0; n_ent_found=0;
  url_params={}
  if args.nmax: url_params['rows'] = args.nmax
  if args.use_compact_associations: url_params['use_compact_associations'] = True
  if args.exclude_automatic_assertions: url_params['exclude_automatic_assertions'] = True
  if args.unselect_evidence: url_params['unselect_evidence'] = True
  url_params_str = ('&'.join(['%s=%s'%(k,str(v)) for k,v in list(url_params.items())]))
  for entid in ids:
    i_ent+=1
    if i_ent<=args.skip: continue
    print('%s:'%entid, file=sys.stderr)
    url_this = (base_url+'/bioentity/disease/%s?%s'%(entid,url_params_str))
    try:
      ent = rest.Utils.GetURL(url_this, parse_json=True)
    except Exception as e:
      if e.args:
        print(str(e.args), file=sys.stderr)
    if ent:
      n_ent_found+=1
      #print(json.dumps(ent, sort_keys=True, indent=2))
      if not tags:
        tags = sorted(ent.keys())
        csvWriter.writerow(tags)
      row_this=[]
      for tag in tags:
        if tag in ent:
          if isinstance(ent[tag], list):
            val = len(ent[tag])
          else:
            val = str(ent[tag])
        else:
          val = ''
        row_this.append(val)
    csvWriter.writerow(row_this)

  print('DEBUG: n_entities = %d'%n_ent_found, file=sys.stderr)


#############################################################################
if __name__=='__main__':
  parser = argparse.ArgumentParser(description='Monarch BioLink REST API client utility')
  ops = ['get_disease', 'get_association']
  parser.add_argument("op", choices=ops, help='operation')
  parser.add_argument("--ifile", help="input file, IDs")
  parser.add_argument("--ids", help="IDs (comma-separated) e.g. DOID:678, OMIM:605543")
  parser.add_argument("--nmax", type=int, help="max results")
  parser.add_argument("--skip", type=int, help="skip results", default=0)
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--use_compact_associations", action="store_true")
  parser.add_argument("--exclude_automatic_assertions", action="store_true")
  parser.add_argument("--unselect_evidence", action="store_true")
  parser.add_argument("--api_host", default=API_HOST)
  parser.add_argument("--api_base_path", default=API_BASE_PATH)
  parser.add_argument("-v", "--verbose", dest="verbose", action="count", default=0)
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  BASE_URL = 'https://'+args.api_host+args.api_base_path

  fout = open(args.ofile, "w+") if args.ofile else sys.stdout

  ids=[];
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      if line.rstrip(): ids.append(line.rstrip())
    fin.close()
  elif args.ids:
    ids = re.split(r'[,\s]+', args.ids)

  if args.op=='get_disease':
    GetDisease(BASE_URL, ids, args, fout)

  elif args.op=='get_association':
    parser.error("Invalid operation: {0}".format(args.op))

  else:
    parser.error("Invalid operation: {0}".format(args.op))




#!/usr/bin/env python
##############################################################################
### pubchem_rdf.py - utility app for the PubChem RDF REST API.
### (Apparently no public Sparql endpoint exists.)
### https://pubchem.ncbi.nlm.nih.gov/rdf
##############################################################################
import sys,os,re,getopt,types,codecs
#
from ...util import rest
#
PROG=os.path.basename(sys.argv[0])
#
API_HOST='rdf.ncbi.nlm.nih.gov'
API_BASE_PATH='/pubchem'
#
OFMTS={
     'CSV': {'ext':'csv',      'mime':'text/csv'},
    'HTML': {'ext':'html',     'mime':'text/html'},
    'JSON': {'ext':'json',     'mime':'application/json'},
     'RDF': {'ext':'rdf',      'mime':'text/rdf'},
  'RDFXML': {'ext':'rdfxml',   'mime':'application/rdf+xml'},
  'TURTLE': {'ext':'ttl',      'mime':'application/x-turtle'}
}
#
#       'NTRIPLES': {'ext':'ntriples',     'mime':'text/plain'},
#             'N3': {'ext':'n3',           'mime':'text/rdf+n3'},
#  'RDFXML-ABBREV': {'ext':'rdfxml-abbrev','mime':'application/rdf+xml+abbrev'},
#############################################################################
### Problem: since each request returns headers, how do we request RDF
### for multiple IDs?  (Hence kludge here.)
#
def GetData(base_uri,pfx,ids,ofmt,fout,verbose):
  i=0;
  for id_query in ids:
    url=(base_uri+'/%s%d.%s'%(pfx,id_query,OFMTS[ofmt.upper()]['ext']))
    res=rest.Utils.GetURL(url,headers={'Accept':'%s'%OFMTS[ofmt.upper()]['mime']},verbose=verbose)
    if not res:
      break
    elif type(res) in types.StringTypes:
      if i>0: #Kludge
        for line in res.splitlines():
          if not line.startswith('@'):
            fout.write(line+'\n')
      else:
        fout.write(res)
      i+=1
    else:
      print >>sys.stderr, 'ERROR: "%s"'%str(res)

#############################################################################
def GetCompound(base_uri,ids,ofmt,fout,verbose):
  GetData(base_uri,'compound/CID',ids,ofmt,fout,verbose)

#############################################################################
def GetSubstance(base_uri,ids,ofmt,fout,verbose):
  GetData(base_uri,'substance/SID',ids,ofmt,fout,verbose)

#############################################################################
def GetProtein(base_uri,ids,ofmt,fout,verbose):
  GetData(base_uri,'protein/GI',ids,ofmt,fout,verbose)

#############################################################################
def GetGene(base_uri,ids,ofmt,fout,verbose):
  GetData(base_uri,'gene/GID',ids,ofmt,fout,verbose)

#############################################################################
def GetBiosystem(base_uri,ids,ofmt,fout,verbose):
  GetData(base_uri,'biosystem/BSID',ids,ofmt,fout,verbose)

#############################################################################
def GetBioassay(base_uri,ids,ofmt,fout,verbose):
  GetData(base_uri,'bioassay/AID',ids,ofmt,fout,verbose)

#############################################################################
def GetEndpoint(base_uri,sids,aids,ofmt,fout,verbose):
  i=0;
  for sid in sids:
    for aid in aids:
      url=(base_uri+'/endpoint/SID%d_AID%d.%s'%(sid,aid,OFMTS[ofmt.upper()]['ext']))
      res=rest.Utils.GetURL(url,headers={'Accept':'%s'%OFMTS[ofmt.upper()]['mime']},verbose=verbose)
      if not res:
        break
      elif type(res) in types.StringTypes:
        if i>0: #Kludge
          for line in res.splitlines():
            if not line.startswith('@'):
              fout.write(line+'\n')
        else:
          fout.write(res)
        i+=1
      else:
        print >>sys.stderr, 'ERROR: "%s"'%str(res)

##############################################################################
if __name__=='__main__':
  ofmt='RDF';
  USAGE='''\
%(PROG)s - (REST client)
required (one of):
        --get_compound ..................(requires CIDs)
        --get_substance .................(requires SIDs)
        --get_protein ...................(requires GIs)
        --get_gene ......................(requires GIDs)
        --get_biosystem .................(requires BSIDs)
        --get_bioassay ..................(requires AIDs)
        --get_endpoint ..................(requires SIDs, AIDs)
options:
	--i IFILE ....................... input IDs file
	--id ID ......................... input ID
	--aidfile AIDFILE ............... input AIDs file
	--aid AID ....................... input AID
	--o OFILE ....................... output file [stdout]
        --ofmt FMT ...................... output formats [%(OFMT)s]
	--api_host HOST ................. [%(API_HOST)s]
	--api_base_path PATH ............ [%(API_BASE_PATH)s]
        --skip N ........................ skip (input IDs)
        --nmax N ........................ max count (input IDs)
        --v ............................. verbose
        --h ............................. this help

Example IDs:
	2244 (compound)
	144206602 (substance)
	124375976 (protein)
	367 (gene)
	82991 (biosystem)
	1117354 (bioassay)

Output formats:
	%(OFMTS)s
'''%{
	'PROG':PROG,
	'API_HOST':API_HOST,
	'API_BASE_PATH':API_BASE_PATH,
	'OFMT':ofmt,
	'OFMTS':('|'.join(OFMTS.keys()))
	}

  def ErrorExit(msg):
    print >>sys.stderr,msg
    sys.exit(1)

  id_query=None; ifile=None;
  aid_query=None; aidfile=None;
  name_query=None;
  ofile='';
  verbose=0;
  get_compound=False;
  get_substance=False;
  get_protein=False;
  get_gene=False;
  get_biosystem=False;
  get_bioassay=False;
  get_endpoint=False;
  skip=None; nmax=None;
  dbusr=None; dbpw=None;
  opts,pargs=getopt.getopt(sys.argv[1:],'',['i=','o=',
	'id=',
	'aid=',
	'aidfile=',
	'ofmt=',
	'get_compound',
	'get_substance',
	'get_protein',
	'get_gene',
	'get_biosystem',
	'get_bioassay',
	'get_endpoint',
	'api_host=','api_base_path=','dbusr=','dbpw=',
	'skip=', 'nmax=','version=','help','v','vv','vvv'])
  if not opts: ErrorExit(USAGE)
  for (opt,val) in opts:
    if opt=='--help': ErrorExit(USAGE)
    elif opt=='--o': ofile=val
    elif opt=='--i': ifile=val
    elif opt=='--id': id_query=int(val)
    elif opt=='--aidfile': aidfile=val
    elif opt=='--aid': aid_query=int(val)
    elif opt=='--ofmt': ofmt=val
    elif opt=='--dbusr': dbusr=val
    elif opt=='--dbpw': DBPW=val
    elif opt=='--get_compound': get_compound=True
    elif opt=='--get_substance': get_substance=True
    elif opt=='--get_protein': get_protein=True
    elif opt=='--get_gene': get_gene=True
    elif opt=='--get_biosystem': get_biosystem=True
    elif opt=='--get_bioassay': get_bioassay=True
    elif opt=='--get_endpoint': get_endpoint=True
    elif opt=='--api_host': API_HOST=val
    elif opt=='--api_base_path': API_BASE_PATH=val
    elif opt=='--skip': skip=int(val)
    elif opt=='--nmax': nmax=int(val)
    elif opt=='--v': verbose=1
    elif opt=='--vv': verbose=2
    elif opt=='--vvv': verbose=3
    else: ErrorExit('Illegal option: %s'%(opt))

  BASE_URL='http://'+API_HOST+API_BASE_PATH

  if ofile:
    fout=open(ofile,"w+")
    #fout=codecs.open(ofile,"w","utf8","replace")
    if not fout: ErrorExit('ERROR: cannot open outfile: %s'%ofile)
  else:
    fout=sys.stdout
    #fout=codecs.getwriter('utf8')(sys.stdout,errors="replace")

  if ofmt.upper() not in OFMTS.keys():
    ErrorExit('ERROR: --ofmt "%s" not allowed.s'%ofmt)

  ids=[]
  if ifile:
    fin=open(ifile)
    if not fin: ErrorExit('ERROR: cannot open ifile: %s'%ifile)
    while True:
      line=fin.readline()
      if not line: break
      try:
        ids.append(int(line.rstrip()))
      except:
        print >>sys.stderr, 'ERROR: bad input ID: %s'%line
        continue
    if verbose:
      print >>sys.stderr, '%s: input IDs: %d'%(PROG,len(ids))
    fin.close()
    id_query=ids[0]
  elif id_query:
    ids=[id_query]

  aids=[]
  if aidfile:
    fin=open(aidfile)
    if not fin: ErrorExit('ERROR: cannot open aidfile: %s'%aidfile)
    while True:
      line=fin.readline()
      if not line: break
      try:
        aids.append(int(line.rstrip()))
      except:
        print >>sys.stderr, 'ERROR: bad input AID: %s'%line
        continue
    if verbose:
      print >>sys.stderr, '%s: input AIDs: %d'%(PROG,len(aids))
    fin.close()
    aid_query=aids[0]
  elif aid_query:
    aids=[aid_query]

  if get_compound:
    if not ids: ErrorExit('ERROR: ID required\n'+USAGE)
    GetCompound(BASE_URL,ids,ofmt,fout,verbose)

  elif get_substance:
    if not ids: ErrorExit('ERROR: ID required\n'+USAGE)
    GetSubstance(BASE_URL,ids,ofmt,fout,verbose)

  elif get_protein:
    if not ids: ErrorExit('ERROR: ID required\n'+USAGE)
    GetProtein(BASE_URL,ids,ofmt,fout,verbose)

  elif get_gene:
    if not ids: ErrorExit('ERROR: ID required\n'+USAGE)
    GetGene(BASE_URL,ids,ofmt,fout,verbose)

  elif get_biosystem:
    if not ids: ErrorExit('ERROR: ID required\n'+USAGE)
    GetBiosystem(BASE_URL,ids,ofmt,fout,verbose)

  elif get_bioassay:
    if not aids: ErrorExit('ERROR: AID required\n'+USAGE)
    GetBioassay(BASE_URL,aids,ofmt,fout,verbose)

  elif get_endpoint:
    if not ids: ErrorExit('ERROR: ID required\n'+USAGE)
    if not aids: ErrorExit('ERROR: AID required\n'+USAGE)
    GetEndpoint(BASE_URL,ids,aids,ofmt,fout,verbose)

  else:
    ErrorExit('ERROR: no operation specified.\n'+USAGE)




#!/usr/bin/env python3
"""
EMBL-EBI Identifiers.org
https://identifiers.org
"""
### 
import sys,os,re,json,argparse,time,logging

from ... import emblebi
#
#############################################################################
if __name__=='__main__':
  epilog='''\
Example IDs: taxonomy:9606
'''
  parser = argparse.ArgumentParser(description='EMBL-EBI Identifiers.org client', epilog=epilog)
  ops = ['resolve',
	'list_namespaces',
	'list_resources',
	'list_institutions',
	'search_namespaces',
	'search_institutions',
	'search_resources',
	]
  parser.add_argument("op", choices=ops, help='OPERATION')
  parser.add_argument("--i", dest="ifile", help="Input IDs")
  parser.add_argument("--o", dest="ofile", help="Output (TSV)")
  parser.add_argument("--ids", help="Input IDs (comma-separated)")
  parser.add_argument("--query", help="Search query.")
  parser.add_argument("--resolver_api_host", default=emblebi.identifiers.RESOLVER_API_HOST)
  parser.add_argument("--resolver_api_base_path", default=emblebi.identifiers.RESOLVER_API_BASE_PATH)
  parser.add_argument("--registry_api_host", default=emblebi.identifiers.REGISTRY_API_HOST)
  parser.add_argument("--registry_api_base_path", default=emblebi.identifiers.REGISTRY_API_BASE_PATH)
  parser.add_argument("--search_logic", choices=("containing", "exact"), default="containing")
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  resolver_api_base_url = 'https://'+args.resolver_api_host+args.resolver_api_base_path
  registry_api_base_url = 'https://'+args.registry_api_host+args.registry_api_base_path

  fout = open(args.ofile, "w+") if args.ofile else sys.stdout

  t0=time.time()

  ids=[]
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.rstrip())
    fin.close()
    logging.info(f"Input IDs: {len(ids)}")
  elif args.ids:
    ids = re.split(r'[,\s]+', args.ids)

  if args.op == "resolve":
    emblebi.identifiers.Utils.Resolve(ids, resolver_api_base_url, fout)

  elif args.op == "list_namespaces":
    emblebi.identifiers.Utils.ListNamespaces(registry_api_base_url, fout)

  elif args.op == "list_resources":
    emblebi.identifiers.Utils.ListResources(registry_api_base_url, fout)

  elif args.op == "list_institutions":
    emblebi.identifiers.Utils.ListInstitutions(registry_api_base_url, fout)

  elif args.op == "search_namespaces":
    emblebi.identifiers.Utils.SearchNamespaces(args.query, args.search_logic, registry_api_base_url, fout)

  elif args.op == "search_institutions":
    emblebi.identifiers.Utils.SearchInstitutions(args.query, args.search_logic, registry_api_base_url, fout)

  else:
    parser.error(f"Invalid operation: {args.op}")

  logging.info(f"Elapsed time: {time.strftime('%Hh:%Mm:%Ss',time.gmtime(time.time()-t0))}")



#!/usr/bin/env python3
"""
	Chem2Bio2RDF query utility.
	All relevant tables assumed to have substring "c2b2r_".
"""
import os,sys,argparse,re,time,logging

from .. import chem2bio2rdf as c2b2r

#############################################################################
if __name__=='__main__':
  parser = argparse.ArgumentParser(description="Chem2Bio2RDF PostgreSql client utility")
  ops = [ "schema", "list_targets", "list_genes", "tablecounts", "get_target", "get_compound" ]
  parser.add_argument("op", choices=ops, help="operation")
  parser.add_argument("--tidfile", dest="tidfile", help="input TID file")
  parser.add_argument("--cidfile", dest="cidfile", help="input CID file")
  parser.add_argument("--tids", help="input TIDs (comma-separated)")
  parser.add_argument("--cids", help="input CIDs (comma-separated)")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--schema_with", help="describe tables w/ SUBSTR in name")
  parser.add_argument("--dbhost")
  parser.add_argument("--dbport", default="5432")
  parser.add_argument("--dbname")
  parser.add_argument("--dbschema", default="public")
  parser.add_argument("--dbusr")
  parser.add_argument("--dbpw")
  parser.add_argument("--param_file", default=os.environ['HOME']+"/.c2b2r.yaml")
  parser.add_argument("-v", "--verbose", dest="verbose", action="count", default=0)
  args = parser.parse_args()

  params = c2b2r.ReadParamFile(args.param_file)
  if args.dbhost: params['DBHOST'] = args.dbhost
  if args.dbport: params['DBPORT'] = args.dbport
  if args.dbname: params['DBNAME'] = args.dbname
  if args.dbusr: params['DBUSR'] = args.dbusr
  if args.dbpw: params['DBPW'] = args.dbpw

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  fout = open(args.ofile, "w+") if args.ofile else sys.stdout

  tids=[]
  if args.tidfile:
    fin = open(args.tidfile)
    while True:
      line = fin.readline()
      if not line: break
      tids.append(line.strip())
  elif args.tids:
    tids = re.split(r'[,\s]+', args.tids)

  cids=[]
  if args.cidfile:
    fin = open(args.cidfile)
    while True:
      line = fin.readline()
      if not line: break
      cids.append(int(line.strip()))
  elif args.cids:
    cids = re.split(r'[,\s]+', args.cids)

  t0=time.time()

  dbcon = c2b2r.Connect(dbhost=params["DBHOST"], dbname=params["DBNAME"], dbusr=params["DBUSR"], dbpw=params["DBPW"])

  if args.op=="schema":
    c2b2r.DescribeSchema(dbcon, args.dbschema, args.schema_with)

  elif args.op=="tablecounts":
    c2b2r.DescribeCounts(dbcon)

  elif args.op=="get_target":
    c2b2r.GetTarget(dbcon, args.dbschema, tids,fout)

  elif args.op=="list_targets":
    c2b2r.ListTargets(dbcon, args.dbschema, fout)

  elif args.op=="list_genes":
    c2b2r.ListGenes(dbcon, args.dbschema, fout)

  elif args.op=="get_compound":
    fout.write('"cid"\t"synonyms"\t"isosmi"\n')
    n_found=0
    for cid in cids:
      ok=c2b2r.GetCompound(dbcon, args.dbschema, cid, fout)
      if ok: n_found+=1
      else: logging.info('CID %d not found'%(cid))
    fout.close()
    logging.info('%d / %d found'%(n_found, len(cids)))

  else:
    parser.error("Invalid operation: {0}".format(args.op))

  dbcon.close()




#!/usr/bin/env python3
"""
Utility for ChEMBL REST API.

Consider using package https://github.com/chembl/chembl_webresource_client.

* https://www.ebi.ac.uk/chembl/api/data/docs
* https://www.ebi.ac.uk/chembl/api/utils/docs
* https://chembl.gitbook.io/chembl-interface-documentation/frequently-asked-questions/general-questions
* https://www.ebi.ac.uk/chembl/faq

ASSAY TYPES:
* Binding (assay_type=B) - Data measuring binding of compound to a molecular target, e.g. Ki, IC50, Kd.
* Functional (assay_type=F) - Data measuring the biological effect of a compound, e.g. %cell death in a cell line, rat weight.
* ADMET (assay_type=A) - ADME data e.g. t1/2, oral bioavailability.
* Toxicity (assay_type=T) - Data measuring toxicity of a compound, e.g., cytotoxicity.
* Physicochemical (assay_type=P) - Assays measuring physicochemical properties of the compounds in the absence of biological material e.g., chemical stability, solubility.
* Unclassified (assay_type=U) - A small proportion of assays cannot be classified into one of the above categories e.g., ratio of binding vs efficacy.

CONFIDENCE_SCORE DESCRIPTION
0 - Default value - Target assignment has yet to be curated
1 - Target assigned is non-molecular
3 - Target assigned is molecular non-protein target
4 - Multiple homologous protein targets may be assigned
5 - Multiple direct protein targets may be assigned
6 - Homologous protein complex subunits assigned
7 - Direct protein complex subunits assigned
8 - Homologous single protein target assigned
9 - Direct single protein target assigned

MOLECULE DEVELOPMENT PHASES:
0 - 
1 - 
2 - 
3 - 
4 - Approved drug?
"""
###
import sys,os,re,json,argparse,time,logging
import pandas as pd
#
from .. import chembl
#
##############################################################################
if __name__=='__main__':
  assay_types = {'B':'Binding', 'F':'Functional', 'A':'ADMET', 'T':'Toxicity', 'P':'Physicochemical', 'U':'Unclassified'}
  epilog='''Assay types: {0}. Example IDs: CHEMBL2 (compound); CHEMBL1642 (compound & drug); CHEMBL240  (target); CHEMBL1824  (target); CHEMBL1217643 (assay); CHEMBL3215220 (assay, PubChem assay 519, NMMLSC FPR); Q12809 (Uniprot)'''.format(str(assay_types))
  parser = argparse.ArgumentParser(description='ChEMBL REST API client', epilog=epilog)
  ops = ["status",
	"list_sources",
	"list_targets",
	"list_assays",
	"list_docs",
	"list_mols",
	"list_drugs",
	"list_drug_indications",
	"list_tissues",
	"list_cells",
	"list_mechanisms",
	"list_organisms",
	"list_protein_classes",
	"search_assays",
	"search_mols_by_name",
	"get_mol",
	"get_mol_by_inchikey",
	"get_target",
	"get_target_components",
	"get_target_by_uniprot",
	"get_assay",
	"get_activity_by_mol",
	"get_activity_by_assay",
	"get_activity_by_target",
	"get_activity_properties",
	"get_document"]
  parser.add_argument("op", choices=ops, help='OPERATION (select one)')
  parser.add_argument("--ids", help="input IDs (e.g. mol, assay, target, document)")
  parser.add_argument("--i", dest="ifile", help="input file, IDs")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--skip", type=int, default=0)
  parser.add_argument("--nmax", type=int, default=None)
  parser.add_argument("--dev_phase", type=int, choices=list(range(5)), default=None, help="molecule development phase")
  parser.add_argument("--assay_source", help="source_id")
  parser.add_argument("--assay_type", help="{0}".format(str(assay_types)))
  parser.add_argument("--pmin", type=float, help="min pChEMBL activity value (9 ~ 1nM *C50)")
  parser.add_argument("--include_phenotypic", action="store_true", help="else pChembl required")
  parser.add_argument("--api_host", default=chembl.API_HOST)
  parser.add_argument("--api_base_path", default=chembl.API_BASE_PATH)
  parser.add_argument("-v","--verbose", action="count", default=0)
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  base_url='https://'+args.api_host+args.api_base_path

  fout = open(args.ofile, 'w') if args.ofile else sys.stdout

  ids=[]
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.rstrip())
    fin.close()
  elif args.ids:
    ids = re.split('[, ]+', args.ids.strip())
  if len(ids)>0: logging.info(f"Input IDs: {len(ids)}")

  if args.op[:3]=="get" and not (args.ifile or args.ids):
    parser.error(f"--i or --ids required for operation {args.op}.")

  if args.op == "status":
    api_ver,db_ver,status = chembl.Status(base_url, fout)

  elif args.op == "list_sources":
    chembl.ListSources(args.api_host, args.api_base_path, fout)

  elif args.op == "list_mechanisms":
    chembl.ListMechanisms(args.api_host, args.api_base_path, fout)

  elif args.op == "list_organisms":
    chembl.ListOrganisms(args.api_host, args.api_base_path, fout)

  elif args.op == "list_protein_classes":
    chembl.ListProteinClasses(args.api_host, args.api_base_path, fout)

  elif args.op == "list_targets":
    chembl.ListTargets(args.skip, args.nmax, args.api_host, args.api_base_path, fout)

  elif args.op == "list_mols":
    chembl.ListMolecules(args.dev_phase, args.skip, args.nmax, args.api_host, args.api_base_path, fout)

  elif args.op == "list_drugs":
    chembl.ListDrugs(args.skip, args.nmax, args.api_host, args.api_base_path, fout)

  elif args.op == "list_drug_indications":
    chembl.ListDrugIndications(args.skip, args.nmax, args.api_host, args.api_base_path, fout)

  elif args.op == "list_docs":
    chembl.ListDocuments(args.skip, args.nmax, args.api_host, args.api_base_path, fout)

  elif args.op == "list_cells":
    chembl.ListCellLines(args.api_host, args.api_base_path, fout)

  elif args.op == "list_tissues":
    chembl.ListTissues(args.api_host, args.api_base_path, fout)

  elif args.op == "list_assays":
    chembl.ListAssays(args.skip, args.nmax, args.api_host, args.api_base_path, fout)

  elif args.op == "get_assay":
    chembl.GetAssay(ids, base_url, fout)

  elif args.op == "get_mol":
    chembl.GetMolecule(ids, base_url, fout)

  elif args.op == "get_mol_by_inchikey":
    chembl.GetMoleculeByInchikey(ids, base_url, fout)

  elif args.op == "get_activity_by_mol":
    chembl.GetActivity(ids, 'molecule', args.pmin, args.skip, args.nmax, args.api_host, args.api_base_path, fout)

  elif args.op == "get_activity_by_assay":
    chembl.GetActivity(ids, 'assay', args.pmin, args.skip, args.nmax, args.api_host, args.api_base_path, fout)

  elif args.op == "get_activity_by_target":
    chembl.GetActivity(ids, 'target', args.pmin, args.skip, args.nmax, args.api_host, args.api_base_path, fout)

  elif args.op == "get_activity_properties":
    chembl.GetActivityProperties(ids, args.skip, args.nmax, base_url, fout)

  elif args.op == "get_target":
    chembl.GetTarget(ids, base_url, fout)

  elif args.op == "get_target_components":
    chembl.GetTargetComponents(ids, args.skip, args.nmax, base_url, fout)

  elif args.op == "get_target_by_uniprot":
    chembl.GetTargetByUniprot(ids, base_url, fout)

  elif args.op == "get_document":
    chembl.GetDocument(ids, args.skip, args.nmax, base_url, fout)

  elif args.op == "search_mols_by_name":
    chembl.SearchMoleculeByName(ids, base_url, fout)

  elif args.op == "search_assays":
    if not (args.assay_source or args.assay_type): parser.error('--assay_source and/or --assay_type required.')
    if args.assay_type and args.assay_type[0].upper() not in ('B', 'F', 'P', 'A', 'U'):
      parser.error(f"Invalid assay type: {args.assay_type}")
    chembl.SearchAssays(args.assay_source, args.assay_type, args.skip, args.nmax, args.api_host, args.api_base_path, fout)

  else:
    parser.error(f"Invalid operation: {args.op}")



#!/usr/bin/env python3
"""
See: http://amp.pharm.mssm.edu/Harmonizome/documentation
"""
###
import sys,os,re,argparse,time,json,logging
#
from ... import maayanlab
#
#
##############################################################################
if __name__=='__main__':
  parser = argparse.ArgumentParser(description='MaayanLab Harmonizome REST API client')
  ops = [ 'get_gene', 'get_gene_associations' ]
  parser.add_argument("op", choices=ops, help='operation')
  parser.add_argument("--i", dest="ifile", help="input IDs")
  parser.add_argument("--ids", help="input IDs (comma-separated)")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--api_host", default=maayanlab.harmonizome.API_HOST)
  parser.add_argument("--api_base_path", default=maayanlab.harmonizome.API_BASE_PATH)
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  base_url = 'https://'+args.api_host+args.api_base_path

  fout = open(args.ofile,"w") if args.ofile else sys.stdout

  t0=time.time()

  ids=[];
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      if line.rstrip(): ids.append(line.rstrip())
    fin.close()
  elif args.ids:
    ids = re.split(r'[,\s]+', args.ids)
  logging.info(f"Input queries: {len(ids)}")

  if args.op == "get_gene":
    maayanlab.harmonizome.Utils.GetGene(ids, base_url, fout)

  elif args.op == "get_gene_associations":
    maayanlab.harmonizome.Utils.GetGeneAssociations(ids, base_url, fout)

  else:
    parser.error(f"Invalid operation: {args.op}")

  logging.info(f"elapsed time: {time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0))}")



#!/usr/bin/env python3
"""
Utility functions for HUGO REST API.
See: http://www.genenames.org/help/rest-web-service-help
"""
import sys,os,re,json,time,logging
import pandas as pd

from ..util import rest
#
API_HOST='rest.genenames.org'
API_BASE_PATH=''
BASE_URL='http://'+API_HOST+API_BASE_PATH
#
OFMTS={'XML':'application/xml','JSON':'application/json'}
#
HEADERS={'Accept':OFMTS['JSON']}
#
##############################################################################
def Info(base_url=BASE_URL, fout=None):
  rval = rest.GetURL(base_url+'/info', headers=HEADERS, parse_json=True)
  logging.info(json.dumps(rval, sort_keys=True, indent=2))

##############################################################################
def ListSearchableFields(base_url=BASE_URL, fout=None):
  rval = rest.GetURL(base_url+'/info', headers=HEADERS, parse_json=True)
  fields = rval['searchableFields'] if 'searchableFields' in rval else None
  fields.sort()
  df = pd.DataFrame({'fields':fields})
  if fout: df.to_csv(fout, "\t", index=False)
  return df

##############################################################################
def GetGenes(qrys, ftypes, skip=0, base_url=BASE_URL, fout=None):
  '''The API can return multiple hits for each query, though should not 
for exact SYMBOL fetch.  One case is for RPS15P5, status "Entry Withdrawn".'''
  n_in=0; n_found=0; n_notfound=0; n_ambig=0; n_gene=0;
  tags=[]; df=None;
  logging.debug('ftypes[{}]: {}'.format(len(ftypes), (','.join(ftypes))))
  for qry in qrys:
    n_in+=1
    if n_in<skip: continue
    logging.debug('{}. query: {}'.format(n_in, qry))
    numFound=0; genes=[];
    ftype_hit=None
    for ftype in ftypes:
      rval = rest.GetURL(base_url+'/fetch/{}/{}'.format(ftype.lower(), qry), headers=HEADERS, parse_json=True)
      if rval is None: continue
      ftype_hit=ftype
      responseHeader = rval['responseHeader']
      logging.debug('response status = {}'.format(responseHeader['status']))
      response = rval['response'] 
      numFound = response['numFound'] if 'numFound' in response else 0
      genes = response['docs']
      if numFound>0: break
    if numFound==0:
      n_notfound+=1
      logging.debug('Not found: {} = {}'.format(ftype_hit, qry))
      continue
    elif numFound>0:
      n_found+=1
    if len(genes)>1:
      logging.debug('Warning: multiple ({}) hits, {} = {} (Duplicate may be status=withdrawn.)'.format(len(genes), ftype_hit, qry))
      n_ambig+=1
    for gene in genes:
      if not tags: tags = list(gene.keys())
      data_this = {'query':qry, 'field':ftype_hit}
      data_this.update({tags[j]:[gene[tags[j]]] for j in range(len(tags))})
      df = pd.concat([df, pd.DataFrame(data_this)])
      n_gene+=1

  logging.info('queries: {}; found: {}; not found: {}; ambiguous: {}; total genes returned: {}'.format(n_in-skip, n_found, n_notfound, n_ambig, n_gene))
  if fout: df.to_csv(fout, "\t", index=False)
  return df

##############################################################################
def SearchGenes(qrys, ftypes, base_url=BASE_URL, fout=None):
  n_in=0; n_found=0; tags=[]; df=None;
  for qry in qrys:
    n_in+=1
    logging.debug('{}. query: {}'.format(n_in, qry))
    found_this=False
    for ftype in ftypes:
      url = (base_url+'/search{}/*{}*'.format((('/'+ftype.lower()) if ftype else ''), qry))
      rval = rest.GetURL(url, headers=HEADERS, parse_json=True)
      responseHeader = rval['responseHeader']
      logging.debug('response status = {}'.format(responseHeader['status']))
      response = rval['response'] 
      numFound = response['numFound'] if 'numFound' in response else 0
      if numFound==0: continue
      found_this=True
      docs = response['docs'] if 'docs' in response else []
      for doc in docs:
        logging.debug(json.dumps(doc, sort_keys=True, indent=2)+'\n')
        if not tags: tags = list(doc.keys())
        data_this = {'query':qry, 'field':ftype}
        data_this.update({tags[j]:[doc[tags[j]]] for j in range(len(tags))})
        df = pd.concat([df, pd.DataFrame(data_this)])
    if found_this: n_found+=1
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info('queries: {}; found: {}'.format(n_in, n_found))
  return df




#!/usr/bin/env python3
"""
Client to HumanBase REST API.
Genome-scale Integrated Analysis of gene Networks in Tissues
GIANT has moved to HumanBase (http://hb.flatironinstitute.org/).
GIANT tissue networks integrate 987 genome-scale datasets, encompassing
~38,000 conditions from ~14,000 publications and include both expression and
interaction measurements.
"""
import sys,os,re,numpy,json,argparse,logging
#
from ..util import rest
#
API_HOST='hb.flatironinstitute.org'
API_BASE_PATH='/api'
#
##############################################################################
def ListDatasets(base_url, fout):
  rval=rest.GetURL(base_url+'/datasets', parse_json=True)
  logging.debug(json.dumps(rval, sort_keys=True, indent=2))
  datasets = rval
  tags=[];
  tags_db=[];
  i_dataset=0;
  for dataset in datasets:
    i_dataset+=1
    if not tags:
      tags=dataset.keys()
      fout.write('%s\n'%('\t'.join(tags)))
    vals=[]
    for tag in tags:
      if tag in dataset:
        vals.append(str(dataset[tag]))
    fout.write('%s\n'%('\t'.join(vals)))
    logging.info('\t%s'%dataset)
  logging.info('n_dataset: %d'%len(datasets))

##############################################################################
def ListDatatypes(base_url, fout):
  rval=rest.GetURL(base_url+'/datatypes', parse_json=True)
  logging.debug(json.dumps(rval, sort_keys=True, indent=2))
  datatypes = rval
  tags=[];
  tags_db=[];
  i_datatype=0;
  for datatype in datatypes:
    i_datatype+=1
    if not tags:
      tags=datatype.keys()
      fout.write('%s\n'%('\t'.join(tags)))
    vals=[]
    for tag in tags:
      if tag in datatype:
        vals.append(str(datatype[tag]))
    fout.write('%s\n'%('\t'.join(vals)))
    logging.info('\t%s'%datatype)
  logging.info('n_datatype: %d'%len(datatypes))

##############################################################################
def ListTerms(base_url, fout):
  rval=rest.GetURL(base_url+'/terms', parse_json=True)
  logging.debug(json.dumps(rval, sort_keys=True, indent=2))
  terms = rval
  tags=[];
  tags_db=[];
  i_term=0;
  for term in terms:
    i_term+=1
    if not tags:
      tags=term.keys()
      fout.write('%s\n'%('\t'.join(tags)))
    vals=[]
    for tag in tags:
      if tag in term:
        vals.append(str(term[tag]))
    fout.write('%s\n'%('\t'.join(vals)))
    logging.info('\t%s'%term)
  logging.info('n_term: %d'%len(terms))

##############################################################################
def ListGenes(base_url, fout):
  rval=rest.GetURL(base_url+'/genes', parse_json=True)
  logging.debug(json.dumps(rval, sort_keys=True, indent=2))
  genes = rval
  for gene in genes:
    fout.write('%s\n'%gene)
    logging.info('\t%s'%gene)
  logging.info('n_gene: %d'%len(genes))

##############################################################################
def GetFuncNet(base_url, tissue, genes, maxsize, prior, minwt, fout):
  url=base_url+'/networks'
  d = {'tissue':tissue, 'genes':('+'.join(genes))}
  if prior: d['prior'] = prior
  if maxsize: d['size'] = maxsize
  rval=rest.PostURL(url,data=d,parse_json=True)

  if not rval: return
  logging.debug(json.dumps(rval,sort_keys=True,indent=2))
  fn = rval
  ver = fn['version'] if fn.has_key('version') else ''
  edges = fn['edges'] if fn.has_key('edges') else []
  nodes = set(genes)
  wts = []
  n_edge=0;
  for edge in edges:
    gene_from, gene_to, weight = edge['source'],edge['target'],edge['weight']
    wt = float(weight)
    if wt<minwt: continue
    fout.write('%s,%s,%.3f\n'%(gene_from, gene_to, wt))
    n_edge+=1
    nodes.add(edge['source'])
    nodes.add(edge['target'])
    wts.append(wt)
  logging.info('weight range: [%.4f,%.4f] ; median: %.4f'%(min(wts),max(wts), numpy.median(wts)))
  logging.info('n_gene: %d (includes query %d)'%(len(nodes),len(genes)))
  logging.info('n_edge: %d'%n_edge)

##############################################################################
def GetFuncNetEvidence(base_url, tissue, source, target, fout):
  url=base_url+'/networks/evidence'
  d = {'tissue':tissue, 'source':source, 'target':target }
  rval=rest.PostURL(url,data=d,parse_json=True)
  logging.debug(json.dumps(rval,sort_keys=True,indent=2))
  tags_dataset = ["dataset", "description", "posterior", "slug", "title", "urltype", "version"]
  fout.write(','.join(tags_dataset)+'\n')
  datasets = rval['datasets'] if rval.has_key('datasets') else []
  for dataset in datasets:
    vals=[]
    for tag in tags_dataset:
      val = dataset[tag] if dataset.has_key(tag) else ''
      vals.append(val)
    fout.write(','.join(vals)+'\n')

  datatypes = rval['datatypes'] if rval.has_key('datatypes') else []
  for datatype in datatypes:
    logging.info('datatype: %s (weight=%s)'%(datatype['name'],datatype['weight']))
 
##############################################################################
if __name__=='__main__':
  FN_SIZE=50; FN_PRIOR=0.1; FN_MIN=0.0; 
  epilog="""\
	--get_funcnet: functional network for tissue, 1+ gene[s];
	--get_evidence: FN evidence for tissue, 2 genes
"""

  parser = argparse.ArgumentParser(description='HumanBase API client; functional genomic networks', epilog=epilog)
  ops=['list_terms', 'list_datasets', 'list_datatypes', 'get_funcnet', 'get_evidence']
  parser.add_argument("op", choices=ops, help='operation')
  parser.add_argument("--tissue", help="(e.g. \"mammary_gland\")")
  parser.add_argument("--gene", help="(e.g. BRCA1)")
  parser.add_argument("--genes", help="gene list (comma-separated)")
  parser.add_argument("--o", dest="ofile", help="output file (TSV)")
  parser.add_argument("--funcnet_size", dest="fn_size", type=int, default=FN_SIZE, help="max genes beyond query")
  parser.add_argument("--funcnet_prior", dest="fn_prior", type=float, default=FN_PRIOR, help="prior probability")
  parser.add_argument("--funcnet_min", dest="fn_min", type=float, default=FN_MIN, help="min probability")
  parser.add_argument("--api_host", default=API_HOST)
  parser.add_argument("--api_base_path", default=API_BASE_PATH)
  parser.add_argument("-v", "--verbose", default=0, action="count")

  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  BASE_URL = 'https://'+args.api_host+args.api_base_path

  if args.ofile:
    fout = open(args.ofile, "w+")
  else:
    fout = sys.stdout

  if args.genes:
    genes=re.split(r'[,\s]+', arg.genes.strip())
  elif args.gene:
    genes=[gene]

  if args.op == 'list_terms':
    ListTerms(BASE_URL, fout)

  elif args.op == 'list_datasets':
    ListDatasets(BASE_URL, fout)

  elif args.op == 'list_datatypes':
    ListDatatypes(BASE_URL, fout)

  elif args.op == 'get_funcnet':
    if not args.tissue and genes: parser.error('tissue and gene[s] required.')
    parser.error('not implemented yet.')
    #GetFuncNet(BASE_URL, args.tissue, genes, args.fn_size, args.fn_prior, args.fn_min, fout)

  elif args.op == 'get_evidence':
    if not args.tissue and genes: parser.error('tissue and gene[s] required.')
    parser.error('not implemented yet.')
    #GetFuncNetEvidence(BASE_URL, args.tissue, genes[0], genes[1], fout)

  else:
    parser.error('Invalid operation: %s'%args.op)




#!/usr/bin/env python3
"""
Access to Ensembl REST API.
http://rest.ensembl.org/documentation/info/lookup
"""
import sys,os,re,argparse,time,json,logging

from .. import ensembl
#
##############################################################################
if __name__=='__main__':
  parser = argparse.ArgumentParser(prog=sys.argv[0], description="Ensembl REST API client", epilog="Example IDs: ENSG00000157764, ENSG00000160785")
  ops = ["list_species", "get_xrefs", "get_info", "show_version"]
  parser.add_argument("op", choices=ops, help='operation')
  parser.add_argument("--ids", help="Ensembl_IDs, comma-separated (ex:ENSG00000000003)")
  parser.add_argument("--i", dest="ifile", help="input file, Ensembl_IDs")
  parser.add_argument("--api_host", default=ensembl.API_HOST)
  parser.add_argument("--api_base_path", default=ensembl.API_BASE_PATH)
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("-v", "--verbose", action="count", default=0)
  parser.add_argument("-q", "--quiet", action="count", default=0)
  args = parser.parse_args()

  # logging.PROGRESS = 15 (custom)
  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>0 else logging.ERROR if args.quiet else 15))

  base_url='http://'+args.api_host+args.api_base_path

  fout = open(args.ofile, "w") if args.ofile else sys.stdout

  t0=time.time()

  if args.ifile:
    fin = open(args.ifile)
    ids=[]
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.strip())
  elif args.ids:
    ids = re.split(r'\s*,\s*', args.ids.strip())
  else:
    ids=[]

  if ids:
    logging.info(f"IDs: {len(ids)}")

  if re.match("^get_", args.op) and not ids:
    parser.error('--i or --ids required.')

  if args.op=='list_species':
    ensembl.ListSpecies(base_url, fout)

  elif args.op=='get_info':
    ensembl.GetInfo(ids, base_url, fout)

  elif args.op=='get_xrefs':
    ensembl.GetXrefs(ids, base_url, fout)

  elif args.op=='show_version':
    ensembl.ShowVersion(base_url, fout)

  else:
    parser.error(f'Invalid operation: {args.op}')

  logging.info(('%s: elapsed time: %s'%(os.path.basename(sys.argv[0]), time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0)))))



#!/usr/bin/env python3
##############################################################################
### Medline Plus Genetics
### https://medlineplus.gov/about/developers/geneticsdatafilesapi/
### https://medlineplus.gov/download/ghr-summaries.xml
##############################################################################
### https://wsearch.nlm.nih.gov/ws/query?db=ghr&term=alzheimer
### https://medlineplus.gov/download/genetics/condition/alzheimer-disease.json
##############################################################################
import sys,os,re,json,argparse,time,logging,requests,urllib.parse
import collections
import pandas as pd
import xmltodict
#
#
API_HOST="wsearch.nlm.nih.gov"
DOWNLOAD_HOST="medlineplus.gov"
API_BASE_PATH="/ws"
DOWNLOAD_BASE_PATH="/download"
#
API_BASE_URL='https://'+API_HOST+API_BASE_PATH
DOWNLOAD_BASE_URL='https://'+DOWNLOAD_HOST+DOWNLOAD_BASE_PATH
#
SUMMARY_URL="https://medlineplus.gov/download/ghr-summaries.xml"
#
##############################################################################
def Search(ids, base_url=API_BASE_URL, fout=None):
  n_out=0; tags=None; df=pd.DataFrame();
  for id_this in ids:
    response = requests.get(f"{base_url}/query?db=ghr&term={urllib.parse.quote(id_this)}")
    rval_dict = xmltodict.parse(response.content)
    logging.debug(json.dumps(rval_dict, indent=2))
    search_results = rval_dict["search_results"]
    count = search_results["@count"]
    query = search_results["query"]
    logging.debug(f"query: {query}; count: {count}")
    results = search_results["result"] if "result" in search_results else []
    for result in results:
      if not tags: tags = list(result.keys())
      df_this = pd.DataFrame({tags[j]:[result[tags[j]]] for j in range(len(tags))})
      if fout is None: df = pd.concat([df, df_this])
      else: df_this.to_csv(fout, "\t", index=False, header=bool(n_out==0))
      n_out += df_this.shape[0]
  logging.info(f"n_out: {n_out}")
  if fout is None: return df

##############################################################################
def ListConditions(summary_url=SUMMARY_URL, fout=None):
  n_out=0; tags=None; df=pd.DataFrame();
  response = requests.get(summary_url)
  logging.debug(f"HTTP Status code: [{response.status_code}]")
  rval_dict = xmltodict.parse(response.content)
  logging.debug(json.dumps(rval_dict, indent=2))
  summaries = rval_dict["summaries"]["health-condition-summary"]
  for summary in summaries:
      if not tags: tags = [tag for tag in summary.keys() if type(summary[tag]) not in (list, dict, collections.OrderedDict)]
      df_this = pd.DataFrame({tags[j]:[summary[tags[j]]] for j in range(len(tags))})
      if fout is None: df = pd.concat([df, df_this])
      else: df_this.to_csv(fout, "\t", index=False, header=bool(n_out==0))
      n_out += df_this.shape[0]
  logging.info(f"n_out: {n_out}")
  if fout is None: return df

##############################################################################
def ListGenes(summary_url=SUMMARY_URL, fout=None):
  n_out=0; tags=None; df=pd.DataFrame();
  response = requests.get(summary_url)
  logging.debug(f"HTTP Status code: [{response.status_code}]")
  rval_dict = xmltodict.parse(response.content)
  logging.debug(json.dumps(rval_dict, indent=2))
  summaries = rval_dict["summaries"]["gene-summary"]
  for summary in summaries:
      if not tags: tags = [tag for tag in summary.keys() if type(summary[tag]) not in (list, dict, collections.OrderedDict)]
      df_this = pd.DataFrame({tags[j]:[summary[tags[j]]] for j in range(len(tags))})
      if fout is None: df = pd.concat([df, df_this])
      else: df_this.to_csv(fout, "\t", index=False, header=bool(n_out==0))
      n_out += df_this.shape[0]
  logging.info(f"n_out: {n_out}")
  if fout is None: return df

##############################################################################
def GetConditionGenes(ids, base_url=DOWNLOAD_BASE_URL, fout=None):
  """IDs must be terms as in condition URLs, e.g. "alzheimer-disease", from
https://medlineplus.gov/download/genetics/condition/alzheimer-disease.json
  """
  n_out=0; tags=None; df=pd.DataFrame();
  for id_this in ids:
    response = requests.get(f"{base_url}/genetics/condition/{id_this}.json")
    if response.status_code != 200:
      logging.error(f"Condition ID not found [{response.status_code}]: '{id_this}'")
      continue
    rval_txt = response.content.decode("utf-8")
    logging.debug(rval_txt)
    rval = json.loads(rval_txt)
    logging.debug(json.dumps(rval, indent=2))
    name = rval["name"]
    ghr_page = rval["ghr_page"]
    genes = rval["related-gene-list"] if "related-gene-list" in rval else []
    for gene in genes:
      df_this = pd.DataFrame({"id":[id_this], "name":[name],
	"gene-symbol":[gene["related-gene"]["gene-symbol"]],
	"ghr-page":[gene["related-gene"]["ghr-page"]]})
      if fout is None: df = pd.concat([df, df_this])
      else: df_this.to_csv(fout, "\t", index=False, header=bool(n_out==0))
      n_out += df_this.shape[0]
  logging.info(f"n_out: {n_out}")
  if fout is None: return df

##############################################################################



#!/usr/bin/env python3
###
"""
Chem2Bio2RDF utility functions.

[ ]	I think we need a function which will add a compound given 
	either
	  *  SMILES and CID 
	or
	  *  InChI and CID 
	The global table is c2b2r_compound.

	With a new ChEMBL compound, the molregno is added separately, and
	linked to the CID.  But this could be done in the same function.
	With SQL it is not so easy to check whether the CID etc. is already
	present.  The function should first check to avoid redundant rows.
	
	Do we also need Inchi2Smiles and Smiles2Inchi functions?
	Can we add the InChI only and get by without smiles?

table: public.c2b2r_compound
	['CID', 'integer']
	['bio2rdf_URI', 'character varying']
	['pubchem_URL', 'character varying']
	['openeye_can_smiles', 'text']
	['openeye_iso_smiles', 'text']
	['std_inchi', 'text']

Also note there is
table: public.c2b2r_compound_new
	['cid', 'integer']
	['bio2rdf_URI', 'character varying']
	['pubchem_URL', 'character varying']
	['openeye_can_smiles', 'text']
	['openeye_iso_smiles', 'text']
	['std_inchi', 'text']
	['gfp', 'bit']
	['gfpbcnt', 'integer']
	??

Problems with c2b2r:
 [ ] Use of schema "public".  Which tables are needed?

"""
###
import os,sys,re,time,math,logging,yaml
import psycopg2,psycopg2.extras

#############################################################################
def ReadParamFile(fparam):
  params={};
  with open(fparam, 'r') as fh:
    for param in yaml.load_all(fh, Loader=yaml.BaseLoader):
      for k,v in param.items():
        params[k] = v
  return params

#############################################################################
def Connect(dbhost, dbname, dbusr, dbpw):
  """Connect to db; specify default cursor type DictCursor."""
  dsn = ("host='%s' dbname='%s' user='%s' password='%s'"%(dbhost, dbname, dbusr, dbpw))
  dbcon = psycopg2.connect(dsn)
  dbcon.cursor_factory = psycopg2.extras.DictCursor
  return dbcon

#############################################################################
def GetCpdBySmi(dbcon, dbschema, smi,iso):
  '''Get compound[s] by smiles; return rows where each row tuple (cid,).'''
  cur=dbcon.cursor()
  if re.search(r'\\',smi): smi=re.sub(r'\\',r'\\\\',smi)
  if iso:
    sql=('SELECT "CID",openeye_can_smiles,openeye_iso_smiles,std_inchi FROM %s.c2b2r_compound WHERE openeye_iso_smiles=gnova.isosmiles(\'%s\')'%(dbschema,smi))
  else:
    sql=('SELECT "CID",openeye_can_smiles,openeye_iso_smiles,std_inchi FROM %s.c2b2r_compound WHERE openeye_iso_smiles=gnova.cansmiles(\'%s\')'%(dbschema,smi))
  cur.execute(sql)
  rows=cur.fetchall()
  cur.close()
  return rows

#############################################################################
def Cids2Smis(dbcon, dbschema, cids):
  '''Return list of SMILES for given list of CIDs.'''
  cur=dbcon.cursor()
  smis=[]
  for cid in cids:
    #sql=('SELECT openeye_can_smiles FROM %s.c2b2r_compound WHERE "CID"=%d'%(dbschema,cid))
    sql=('SELECT openeye_iso_smiles FROM %s.c2b2r_compound WHERE "CID"=%d'%(dbschema,cid))
    #sql=('SELECT openeye_iso_smiles FROM %s.c2b2r_compound_new WHERE "cid"=%d'%(dbschema,cid))
    cur.execute(sql)
    rows=cur.fetchall()
    if not rows: smis.append(None)
    elif len(rows[0])<1:  smis.append(None)
    else: smis.append(rows[0][0])
  cur.close()
  return smis

#############################################################################
def DescribeSchema(dbcon, dbschema, substr):
  '''Return human readable text describing the schema.'''
  cur=dbcon.cursor()
  if substr:
    sql=("SELECT table_name FROM information_schema.tables WHERE table_schema='%s' AND table_name ILIKE '%%c2b2r_%%' AND table_name ILIKE '%%%s%%'"%(dbschema,substr))
  else:
    sql=("SELECT table_name FROM information_schema.tables WHERE table_schema='%s' AND table_name ILIKE '%%c2b2r_%%'"%dbschema)
  cur.execute(sql)
  rows=cur.fetchall()	##data rows
  n_table=len(rows)
  for row in rows:
    tablename=row[0]
    sql=("SELECT column_name,data_type FROM information_schema.columns WHERE table_schema='%s' AND table_name = '%s'"%(dbschema,tablename))
    cur.execute(sql)
    rows=cur.fetchall()	##data rows
    print("table: %s.%s"%(dbschema,tablename))
    for row in rows:
      print("\t%s"%str(row))
  cur.close()
  logging.info("table count: %d"%n_table)

#############################################################################
def DescribeCounts(dbcon, dbschema="public"):
  '''Return human readable text listing the table rowcounts.'''
  cur=dbcon.cursor()
  sql=("SELECT table_name FROM information_schema.tables WHERE table_schema='%s' AND table_name ILIKE '%%c2b2r_%%' ORDER BY table_name"%dbschema)
  cur.execute(sql)
  rows=cur.fetchall()
  print("table rowcounts:")
  for row in rows:
    tablename=row[0]
    sql=("SELECT count(*) FROM %s.\"%s\""%(dbschema,tablename))
    cur.execute(sql)
    rows=cur.fetchall()
    print("%-40s: %9d"%(tablename,rows[0][0]))
  cur.close()
 
#############################################################################
def AddNewCID(dbcon, dbschema, cid,smi,inchi):
  cur=dbcon.cursor()
  smi=re.sub(r'\\',r"'||E'\\\\'||'",smi)
  sql=("INSERT INTO %s.c2b2r_compound (\"CID\",openeye_can_smiles,openeye_iso_smiles,std_inchi) VALUES (%d,gnova.cansmiles('%s'),gnova.isosmiles('%s'),'%s')"%(dbschema,cid,smi,smi,inchi))
  ok=cur.execute(sql)
  cur.close()
  return ok

#############################################################################
def GetTarget(dbcon, dbschema, tids,fout):
  cur=dbcon.cursor()
  sql='''\
SELECT DISTINCT
	protein_accession,
	db_source,
	pref_name AS name
FROM
	%(DBSCHEMA)s.c2b2r_chembl_08_target_dictionary
WHERE
	target_type = 'PROTEIN'
	AND organism = 'Homo sapiens'
	AND db_source = 'SWISS-PROT'
	AND protein_accession IN ( %(TIDS)s )
ORDER BY protein_accession ASC
'''%{'DBSCHEMA':dbschema,'TIDS':','.join(map(lambda x:("'%s'"%x),tids))}
  ok=cur.execute(sql)
  rows=cur.fetchall()
  fout.write('"protein_accession","db_source","name"\n')
  for row in rows:
    fout.write('"%s","%s","%s"\n'%(row[0],row[1],row[2]))
  cur.close()
  return ok

#############################################################################
def ListTargets(dbcon, dbschema, fout):
  cur=dbcon.cursor()
  sql='''\
SELECT DISTINCT
	tid,
	protein_accession,
	chembl_id,
	tax_id,
	target_type,
	organism,
	db_source,
	gene_symbol,
	pref_name
FROM
	public.c2b2r_chembl_08_target_dictionary
WHERE
	target_type = 'PROTEIN'
'''%{'DBSCHEMA':dbschema}
  fout.write('"tid","protein_accession","chembl_id","tax_id","target_type","organism","db_source","gene_symbol","pref_name"\n')
  ok=cur.execute(sql)
  rows=cur.fetchall()
  for row in rows:
    for i,val in enumerate(row):
      if val==None: val=''
      fout.write('"%s"'%(val))
      if i<(len(row)-1): fout.write(',')
    fout.write('\n')
    fout.flush()
  cur.close()

#############################################################################
def ListGenes(dbcon, dbschema, fout):
  cur=dbcon.cursor()
  sql='''\
SELECT DISTINCT
	tid,
	protein_accession,
	chembl_id,
	tax_id,
	target_type,
	organism,
	db_source,
	gene_symbol,
	pref_name
FROM
	public.c2b2r_chembl_08_target_dictionary
WHERE
	target_type = 'PROTEIN'
ORDER BY
	gene_symbol ASC
'''%{'DBSCHEMA':dbschema}
  fout.write('"tid","protein_accession","chembl_id","tax_id","target_type","organism","db_source","gene_symbol","pref_name"\n')
  ok=cur.execute(sql)
  rows=cur.fetchall()
  for row in rows:
    for i,val in enumerate(row):
      if val==None: val=''
      fout.write('"%s"'%(val))
      if i<(len(row)-1): fout.write(',')
    fout.write('\n')
    fout.flush()
  cur.close()

#############################################################################
### Very few synonyms in c2b2r_compound_synonym (506).
### But!  87615736 synonyms in:
### table: public.pubchem_synonym
###        ['ccid_synid', 'bigint']
###        ['cid', 'character varying']
###        ['synonym', 'text']
#############################################################################
def GetCompound(dbcon, dbschema, cid,fout):
  cur=dbcon.cursor()
  sql='''\
SELECT DISTINCT
	cpd.cid,
	syn.synonym AS name,
	cpd.openeye_iso_smiles AS isosmi
FROM
	%(DBSCHEMA)s.c2b2r_compound_new AS cpd
LEFT OUTER JOIN
	pubchem_synonym AS syn ON (syn.cid::INT = cpd.cid)
WHERE
	cpd.cid = %(CID)d
ORDER BY cid, name
'''%{'DBSCHEMA':dbschema,'CID':cid}
  cur.execute(sql)
  rows=cur.fetchall()
  if not rows: return False
  names=[]
  for row in rows:
    name=row[1]
    isosmi=row[2]
    if name: names.append(name)

  names = SortCompoundNamesForHumans(names)
  for i,name in enumerate(names):
    if i==10: break
    logging.debug('"%s"'%(name))
  if len(names)>10: names=names[:10]
  names_str=';'.join(names)
  fout.write('%d,"%s","%s"\n'%(cid,names_str,isosmi))
  fout.flush()
  cur.close()
  return True

#############################################################################
### Heuristic for human comprehensibility.
#############################################################################
def SortCompoundNamesForHumans(names):
  names_scored={n:0 for n in names}
  pat_proper = re.compile(r'^[A-Z][a-z]+$')
  pat_text = re.compile(r'^[A-Za-z ]+$')
  for name in names:
    if pat_proper.match(name): names_scored[name]+=100
    elif pat_text.match(name): names_scored[name]+=50
    elif re.match(r'^[A-z][A-z][A-z][A-z][A-z][A-z][A-z].*$',name): names_scored[name]+=10
    if re.search(r'[\[\]\{\}]',name): names_scored[name]-=50
    if re.search(r'[_/\(\)]',name): names_scored[name]-=10
    if re.search(r'\d',name): names_scored[name]-=10
    names_scored[name] -= math.fabs(7-len(name))    #7 is optimal!

  names_ranked = [[score,name] for name,score in names_scored.items()]
  names_ranked.sort()
  names_ranked.reverse()
  #for score,name in names_ranked:
  #  logging.debug('"%s" (%d)'%(name,score))
  return [name for score,name in names_ranked]

#############################################################################



#!/usr/bin/env python3
"""
Reactome REST API utilities.
https://reactome.org/ContentService/
https://reactome.org/ContentService/data/pathways/top/9606

https://reactome.org/documentation/data-model

"Life on the cellular level is a network of molecular interactions. Molecules are synthesized and degraded, undergo a bewildering array of temporary and permanent modifications, are transported from one location to another, and form complexes with other molecules. Reactome represents all of this complexity as reactions in which input physical entities are converted to output entities."

"PhysicalEntities include individual molecules, multi-molecular complexes, and sets of molecules or complexes grouped together on the basis of shared characteristics. Molecules are further classified as genome encoded (DNA, RNA, and proteins) or not (all others). Attributes of a PhysicalEntity instance capture the chemical structure of an entity, including any covalent modifications in the case of a macromolecule, and its subcellular localization."
"""
import sys,os,re,json,time,logging
import pandas as pd
#
from ..util import rest
#
API_HOST='reactome.org'
API_BASE_PATH='/ContentService'
BASE_URL='https://'+API_HOST+API_BASE_PATH
#
HEADERS={'Content-type':'text/plain', 'Accept':'application/json'}
#
##############################################################################
def DBInfo(base_url=BASE_URL, fout=None):
  name = rest.GetURL(base_url+'/data/database/name')
  version = rest.GetURL(base_url+'/data/database/version')
  df = pd.DataFrame({'param':['name', 'version'], 'value':[name, version]});
  if fout: df.to_csv(fout, "\t", index=False)
  return df

##############################################################################
def ListToplevelPathways(base_url=BASE_URL, fout=None):
  tags=None; species="9606"; df=pd.DataFrame();
  pathways = rest.GetURL(base_url+f'/data/pathways/top/{species}', parse_json=True)
  for pathway in pathways:
    if not tags: tags = list(pathway.keys())
    df = pd.concat([df, pd.DataFrame({tags[j]:[pathway[tags[j]]] if tags[j] in pathway else [''] for j in range(len(tags))})])
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info('Top-level pathways: {}'.format(df.shape[0]))
  return df

##############################################################################
def ListDiseases(base_url=BASE_URL, fout=None):
  tags=[]; df=pd.DataFrame();
  rval = rest.GetURL(base_url+'/data/diseases', parse_json=True)
  diseases = rval
  for disease in diseases:
    if not tags: tags = list(disease.keys())
    df = pd.concat([df, pd.DataFrame({tags[j]:[disease[tags[j]]] if tags[j] in disease else [''] for j in range(len(tags))})])
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info('n_diseases: {}'.format(df.shape[0]))
  return df

##############################################################################
def QueryEntry(ids, base_url=BASE_URL, fout=None):
  """Stable or database IDs required, such as for disease, or pathway."""
  tags=None; df=pd.DataFrame(); classNames=set();
  for id_this in ids:
    rval = rest.GetURL(base_url+f'/data/query/{id_this}', parse_json=True)
    logging.debug(json.dumps(rval, sort_keys=True, indent=2)+'\n')
    ent = rval
    if not tags: tags = list(ent.keys())
    if 'className' in ent: classNames.add(ent['className'])
    df = pd.concat([df, pd.DataFrame({tags[j]:[ent[tags[j]]] if tags[j] in ent else [''] for j in range(len(tags))})])
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info('n_out {}: {}'.format((list(classNames)), df.shape[0]))
  return df

##############################################################################
def GetInteractors(ids, base_url=BASE_URL, fout=None):
  """IDs may be UniProt accessions."""
  tags=None; df=pd.DataFrame(); classNames=set();
  for id_this in ids:
    rval = rest.GetURL(base_url+f'/interactors/static/molecule/{id_this}/details', parse_json=True)
    logging.debug(json.dumps(rval, sort_keys=True, indent=2)+'\n')
    ents = rval["entities"] if "entities" in rval else []
    for ent in ents:
      intrs = ent["interactors"] if "interactors" in ents else []
      for intr in intrs:
        if not tags: tags = list(intr.keys())
        if 'className' in intr: classNames.add(intr['className'])
        df = pd.concat([df, pd.DataFrame({tags[j]:[intr[tags[j]]] if tags[j] in intr else [''] for j in range(len(tags))})])
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info('n_out {}: {}'.format((list(classNames)), df.shape[0]))
  return df

##############################################################################
def ListCompounds(base_url=BASE_URL, fout=None):
  """NOT WORKING."""
  tags=[]; df=pd.DataFrame();
  rval = rest.GetURL(base_url+'/getReferenceMolecules', parse_json=True)
  mols = rval
  for mol in mols:
    if not tags: tags = list(mol.keys())
    df = pd.concat([df, pd.DataFrame({tags[j]:[mol[tags[j]]] for j in range(len(tags))})])
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info('n_mols: {}'.format(df.shape[0]))
  return df

##############################################################################
def GetPathwaysForEntities(ids, base_url=BASE_URL, fout=None):
  """NOT WORKING."""
  tags=[]; df=pd.DataFrame();
  d=('ID='+(','.join(ids))) ##plain text POST body, e.g. "ID=170075,176374,68557"
  rval = rest.PostURL(base_url+'/pathwaysForEntities', data=d, headers=HEADERS, parse_json=True)
  logging.debug(json.dumps(rval, sort_keys=True, indent=2)+'\n')
  pathways = rval
  for pathway in pathways:
    if not tags: tags = list(set(pathway.keys()))
    df = pd.concat([df, pd.DataFrame({tags[j]:[pathway[tags[j]]] for j in range(len(tags))})])
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info('n_in: {}; n_out: {}'.format(len(ids), df.shape[0]))
  return df

##############################################################################
def GetPathwaysForGenes(ids, base_url=BASE_URL, fout=None):
  """NOT WORKING."""
  tags=[]; df=pd.DataFrame();
  d=(','.join(ids)) ##plain text POST body
  rval = rest.PostURL(base_url+'/queryHitPathways', data=d, headers=HEADERS, parse_json=True)
  logging.debug(json.dumps(rval, sort_keys=True, indent=2)+'\n')
  pathways = rval
  for pathway in pathways:
    if not tags: tags = list(set(pathway.keys()))
    df = pd.concat([df, pd.DataFrame({tags[j]:[pathway[tags[j]]] for j in range(len(tags))})])
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info('n_in: {}; n_out: {}'.format(len(ids), df.shape[0]))
  return df

##############################################################################
def GetPathwayParticipants(id_query, base_url=BASE_URL, fout=None):
  """NOT WORKING."""
  tags=[]; df=pd.DataFrame();
  rval = rest.GetURL(base_url+'/pathwayParticipants/'+id_query, parse_json=True)
  logging.debug(json.dumps(rval, sort_keys=True, indent=2)+'\n')
  parts = rval
  for part in parts:
    if not tags: tags = list(set(part.keys()))
    df = pd.concat([df, pd.DataFrame({tags[j]:[part[tags[j]]] for j in range(len(tags))})])
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info('n_in: {}; n_out: {}'.format(len(ids), df.shape[0]))
  return df

##############################################################################



#!/usr/bin/env python3
"""
See: https://api.jensenlab.org/About
"""
###
import sys,os,re,argparse,time,json,logging
#
from .. import jensenlab
#
##############################################################################
if __name__=='__main__':
  CHANNELS= ['Knowledge', 'Experiments', 'Textmining', 'All']
  parser = argparse.ArgumentParser(description='JensenLab REST API client')
  ops = ['get_disease_genes', 'get_comention_genes' ]
  parser.add_argument("op", choices=ops, help='OPERATION')
  parser.add_argument("--i", dest="ifile", help="input IDs (for diseases should be DOIDs, e.g. \"DOID:10652\"")
  parser.add_argument("--channel", choices=CHANNELS, default="Textmining", help="source channel")
  parser.add_argument("--ids", help="input IDs (comma-separated)")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--nmax", type=int, default=100, help="max hits")
  parser.add_argument("--api_host", default=jensenlab.API_HOST)
  parser.add_argument("--api_base_path", default=jensenlab.API_BASE_PATH)
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  base_url = 'https://'+args.api_host+args.api_base_path

  fout = open(args.ofile,"w") if args.ofile else sys.stdout

  t0 = time.time()

  ids=[];
  if args.ifile:
    with open(args.ifile) as fin:
      while True:
        line = fin.readline()
        if not line: break
        if line.rstrip(): ids.append(line.rstrip())
  elif args.ids:
    ids = re.split(r'[,\s]+', args.ids)
  logging.info('Input queries: %d'%(len(ids)))

  if args.op == "get_disease_genes":
    jensenlab.GetDiseaseGenes(args.channel, ids, args.nmax, base_url, fout)

  elif args.op == "get_comention_genes":
    jensenlab.GetPubmedComentionGenes(ids, base_url, fout)

  else:
    parser.error(f"Invalid operation: {args.op}")

  logging.info(('elapsed time: %s'%(time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0)))))
